from __future__ import annotations

import logging
import math
import os
from typing import List

import numpy as np
from jinja2 import ChoiceLoader, Environment, PackageLoader

from scitbx.array_family import flex

from dials.algorithms.clustering import plots as cluster_plotter
from dials.algorithms.clustering.unit_cell import (
    Cluster,
    ClusteringResult,
    cluster_unit_cells,
)
from dials.util import tabulate

logger = logging.getLogger("dials.algorithms.indexing.ssx.analysis")


def generate_html_report(plots: dict, filename: str | os.PathLike) -> None:
    loader = ChoiceLoader(
        [
            PackageLoader("dials", "templates"),
            PackageLoader("dials", "static", encoding="utf-8"),
        ]
    )
    env = Environment(loader=loader)
    template = env.get_template("simple_report.html")
    html = template.render(
        page_title="DIALS SSX indexing report",
        panel_title="Indexing plots",
        graphs=plots,
    )
    with open(filename, "wb") as f:
        f.write(html.encode("utf-8", "xmlcharrefreplace"))


def make_summary_table(results_summary: dict) -> tabulate:
    # make a summary table
    overall_summary_header = [
        "Image",
        "expt_id",
        "n_indexed",
        "RMSD X",
        "RMSD Y",
        "RMSD dPsi",
    ]

    rows = []
    total = 0
    if any(len(v) > 1 for v in results_summary.values()):
        show_lattices = True
        overall_summary_header.insert(1, "lattice")
    else:
        show_lattices = False
    for k in sorted(results_summary.keys()):
        for j, cryst in enumerate(results_summary[k]):
            if not cryst["n_indexed"] or not cryst["n_strong"]:
                continue
            n_idx, n_strong = (cryst["n_indexed"], cryst["n_strong"])
            frac_idx = f"{n_idx}/{n_strong} ({100*n_idx/n_strong:2.1f}%)"
            row = [
                cryst["Image"],
                str(total),
                frac_idx,
                cryst["RMSD_X"],
                cryst["RMSD_Y"],
                cryst["RMSD_dPsi"],
            ]
            if show_lattices:
                row.insert(1, j + 1)
            rows.append(row)
            total += 1

    summary_table = tabulate(rows, overall_summary_header)
    return summary_table


def combine_results_dicts(results_summaries: List[dict]) -> dict:
    """For a list of dictionaries, each with keys 0..n-1,
    combine into a single dictionary with keys 0..ntot-1"""
    combined_summary = {}
    n_overall = 0
    for d in results_summaries:
        n_this = len(d)
        for i in range(n_this):
            combined_summary[i + n_overall] = d.pop(i)
        n_overall += n_this
    return combined_summary


def make_cluster_plots(large_clusters: List[Cluster]) -> dict:
    cluster_plots = {}
    for n, cluster in enumerate(large_clusters):
        uc_params = [flex.double() for i in range(6)]
        for cs in cluster.crystal_symmetries:
            ucp = cs.unit_cell().parameters()
            for i in range(6):
                uc_params[i].append(ucp[i])
        d_this = cluster_plotter.plot_uc_histograms(uc_params)
        d_this["uc_scatter"]["layout"]["title"] += f" cluster {n+1}"
        d_this["uc_hist"]["layout"]["title"] += f" cluster {n+1}"
        if "uc_angle_hist" in d_this:
            d_this["uc_angle_hist"]["layout"]["title"] += f" cluster {n+1}"
        d_this[f"uc_scatter_{n}"] = d_this.pop("uc_scatter")
        d_this[f"uc_hist_{n}"] = d_this.pop("uc_hist")
        if "uc_angle_hist" in d_this:
            d_this[f"uc_angle_hist_{n}"] = d_this.pop("uc_angle_hist")
        cluster_plots.update(d_this)
    return cluster_plots


def report_on_crystal_clusters(crystal_symmetries, make_plots=True):
    clustering = cluster_unit_cells(
        crystal_symmetries,
        threshold=5000,
    )
    cluster_plots = {}
    large_clusters = []
    min_cluster_pc = 5
    if not clustering:  # could happen e.g. if only one unit cell
        large_clusters = [Cluster(crystal_symmetries)]
    else:
        threshold = math.floor((min_cluster_pc / 100) * len(crystal_symmetries))
        large_clusters = [c for c in clustering.clusters if len(c) > threshold]
        large_clusters.sort(key=len, reverse=True)

    if large_clusters:
        logger.info(
            f"""
Unit cell clustering analysis, clusters with >{min_cluster_pc}% of the number of crystals indexed
{ClusteringResult(large_clusters)}
"""
        )
        if make_plots:
            cluster_plots = make_cluster_plots(large_clusters)
    else:
        logger.info(
            f"No clusters found with >{min_cluster_pc}% of the number of crystals."
        )
    return cluster_plots, large_clusters


def generate_plots(summary_data: dict) -> dict:
    """Generate indexing plots from the summary data from index_all_concurrent"""
    # n_indexed_arrays are cumulative n_indexed for nth lattice
    n_indexed_arrays = [np.zeros(len(summary_data))]
    rmsd_x_arrays = [np.zeros(len(summary_data))]
    rmsd_y_arrays = [np.zeros(len(summary_data))]
    rmsd_z_arrays = [np.zeros(len(summary_data))]
    n_total_indexed = np.zeros(len(summary_data))
    n_strong_array = np.zeros(len(summary_data))
    n_lattices = 1

    sorted_keys = sorted(summary_data.keys())
    images = np.array(sorted_keys)

    for i, k in enumerate(sorted_keys):
        n_lattices_this = len(summary_data[k])
        n_strong_array[i] = summary_data[k][0]["n_strong"]
        for j, cryst in enumerate(summary_data[k]):
            if not cryst["n_indexed"]:
                continue
            if n_lattices_this > n_lattices:
                for _ in range(n_lattices_this - n_lattices):
                    n_indexed_arrays.append(np.zeros(len(summary_data)))
                    rmsd_x_arrays.append(np.zeros(len(summary_data)))
                    rmsd_y_arrays.append(np.zeros(len(summary_data)))
                    rmsd_z_arrays.append(np.zeros(len(summary_data)))
                n_lattices = n_lattices_this
            n_indexed_arrays[j][i] = cryst["n_indexed"]
            rmsd_x_arrays[j][i] = cryst["RMSD_X"]
            rmsd_y_arrays[j][i] = cryst["RMSD_Y"]
            rmsd_z_arrays[j][i] = cryst["RMSD_dPsi"]
            n_total_indexed[i] += cryst["n_indexed"]

    n_indexed_data = [
        {
            "x": images.tolist(),
            "y": n_indexed_arrays[0].tolist(),
            "type": "scatter",
            "mode": "markers",
            "name": "N indexed",
        },
    ]
    rmsd_data = [
        {
            "x": images[rmsd_x_arrays[0] > 0].tolist(),
            "y": rmsd_x_arrays[0][rmsd_x_arrays[0] > 0].tolist(),
            "type": "scatter",
            "mode": "markers",
            "name": "RMSD X",
        },
        {
            "x": images[rmsd_y_arrays[0] > 0].tolist(),
            "y": rmsd_y_arrays[0][rmsd_y_arrays[0] > 0].tolist(),
            "type": "scatter",
            "mode": "markers",
            "name": "RMSD Y",
        },
    ]
    rmsdz_data = [
        {
            "x": images[rmsd_z_arrays[0] > 0].tolist(),
            "y": rmsd_z_arrays[0][rmsd_z_arrays[0] > 0].tolist(),
            "type": "scatter",
            "mode": "markers",
            "name": "RMSD dPsi",
        },
    ]
    if n_lattices > 1:
        n_indexed_data[0]["name"] += " (lattice 1)"
        rmsd_data[0]["name"] += " (lattice 1)"
        rmsd_data[1]["name"] += " (lattice 1)"
        rmsdz_data[0]["name"] += " (lattice 1)"
        for i, arr in enumerate(n_indexed_arrays[1:]):
            sub_images = images[arr > 0]
            sub_data = arr[arr > 0]
            n_indexed_data.append(
                {
                    "x": sub_images.tolist(),
                    "y": sub_data.tolist(),
                    "type": "scatter",
                    "mode": "markers",
                    "name": f"N indexed (lattice {i+2})",
                }
            )
        for i, arr in enumerate(rmsd_x_arrays[1:]):
            sub_images = images[arr > 0]
            sub_data_x = arr[arr > 0]
            sub_data_y = rmsd_y_arrays[i + 1][arr > 0]
            rmsd_data.append(
                {
                    "x": sub_images.tolist(),
                    "y": sub_data_x.tolist(),
                    "type": "scatter",
                    "mode": "markers",
                    "name": f"RMSD X (lattice {i+2})",
                },
            )
            rmsd_data.append(
                {
                    "x": sub_images.tolist(),
                    "y": sub_data_y.tolist(),
                    "type": "scatter",
                    "mode": "markers",
                    "name": f"RMSD Y (lattice {i+2})",
                },
            )
        for i, arr in enumerate(rmsd_z_arrays[1:]):
            sub_images = images[arr > 0]
            sub_data = arr[arr > 0]
            rmsdz_data.append(
                {
                    "x": sub_images.tolist(),
                    "y": sub_data.tolist(),
                    "type": "scatter",
                    "mode": "markers",
                    "name": f"RMSD dPsi (lattice {i+2})",
                },
            )
    percent_indexed = np.zeros(shape=(n_total_indexed.size,))
    sel = n_strong_array > 0
    sel_n_tot = n_total_indexed[sel]
    sel_n_strong = n_strong_array[sel]
    percent_indexed[sel] = 100 * sel_n_tot / sel_n_strong
    images = images.tolist()
    n_indexed_data.append(
        {
            "x": images,
            "y": n_strong_array.tolist(),
            "type": "scatter",
            "mode": "markers",
            "name": "N strong",
        },
    )

    percent_bins = np.linspace(0, 100, 51)
    percent_hist = np.histogram(percent_indexed, percent_bins)[0]

    def _generate_hist_data(rmsd_arrays, step=0.01):
        all_rmsd = np.concatenate(rmsd_arrays)
        all_rmsd = all_rmsd[all_rmsd > 0]
        if len(all_rmsd) == 1:
            return (
                np.array([0, 1, 0]),
                np.array([all_rmsd[0] - step, all_rmsd[0], all_rmsd[0] + step]),
            )
        mult = int(1 / 0.01)
        start = math.floor(np.min(all_rmsd) * mult) / mult
        stop = math.ceil(np.max(all_rmsd) * mult) / mult
        nbins = max(1, int((stop - start) / step))
        hist, bin_edges = np.histogram(
            all_rmsd,
            bins=nbins,
            range=(start, stop),
        )
        bin_centers = bin_edges[:-1] + np.diff(bin_edges) / 2
        return hist, bin_centers

    hist_x, bin_centers_x = _generate_hist_data(rmsd_x_arrays)
    hist_y, bin_centers_y = _generate_hist_data(rmsd_y_arrays)
    hist_z, bin_centers_z = _generate_hist_data(rmsd_z_arrays, 0.001)

    plots = {
        "n_indexed": {
            "data": n_indexed_data,
            "layout": {
                "title": "Number of indexed reflections per image",
                "xaxis": {"title": "image number"},
                "yaxis": {"title": "N reflections"},
            },
        },
        "percent_indexed": {
            "data": [
                {
                    "x": images,
                    "y": percent_indexed.tolist(),
                    "type": "scatter",
                    "mode": "markers",
                    "name": "Percentage of strong spots indexed",
                }
            ],
            "layout": {
                "title": "Percentage of strong spots indexed per image",
                "xaxis": {"title": "image number"},
                "yaxis": {"title": "Percentage"},
            },
        },
        "percent_indexed_hist": {
            "data": [
                {
                    "x": percent_bins.tolist(),
                    "y": percent_hist.tolist(),
                    "type": "bar",
                }
            ],
            "layout": {
                "title": "Distribution of percentage indexed",
                "xaxis": {"title": "Percentage indexed"},
                "yaxis": {"title": "Number of images"},
                "bargap": 0,
            },
        },
        "rmsds": {
            "data": rmsd_data,
            "layout": {
                "title": "RMSDs (x, y) per image",
                "xaxis": {"title": "image number"},
                "yaxis": {"title": "RMSD (px)"},
            },
        },
        "rmsdz": {
            "data": rmsdz_data,
            "layout": {
                "title": "RMSD (dPsi) per image",
                "xaxis": {"title": "image number"},
                "yaxis": {"title": "RMSD dPsi (deg)"},
            },
        },
        "rmsdxy_hist": {
            "data": [
                {
                    "x": bin_centers_x.tolist(),
                    "y": hist_x.tolist(),
                    "type": "bar",
                    "name": "RMSD X",
                    "opacity": 0.6,
                },
                {
                    "x": bin_centers_y.tolist(),
                    "y": hist_y.tolist(),
                    "type": "bar",
                    "name": "RMSD Y",
                    "opacity": 0.6,
                },
            ],
            "layout": {
                "title": "Distribution of RMSDs (x, y)",
                "xaxis": {"title": "RMSD (px)"},
                "yaxis": {"title": "Number of images"},
                "bargap": 0,
                "barmode": "overlay",
            },
        },
        "rmsdz_hist": {
            "data": [
                {
                    "x": bin_centers_z.tolist(),
                    "y": hist_z.tolist(),
                    "type": "bar",
                    "name": "RMSD dPsi",
                },
            ],
            "layout": {
                "title": "Distribution of RMSDs (dPsi)",
                "xaxis": {"title": "RMSD dPsi (deg)"},
                "yaxis": {"title": "Number of images"},
                "bargap": 0,
            },
        },
    }
    return plots



from __future__ import annotations

import os
import shutil

import procrunner

from dxtbx.serialize import load

from dials.command_line.show import model_connectivity, run


def test_dials_show(dials_regression):
    path = os.path.join(dials_regression, "experiment_test_data", "experiment_1.json")
    result = procrunner.run(
        ["dials.show", path], environment_override={"DIALS_NOBANNER": "1"}
    )
    assert not result.returncode and not result.stderr
    output = result.stdout.decode("latin-1")
    output = [_f for _f in (s.rstrip() for s in output.split("\n")) if _f]
    assert (
        "\n".join(output[6:])
        == """
Detector:
Panel:
  name: Panel
  type: SENSOR_PAD
  identifier:
  pixel_size:{0.172,0.172}
  image_size: {2463,2527}
  trusted_range: {-1,495976}
  thickness: 0
  material:
  mu: 0
  gain: 1
  pedestal: 0
  fast_axis: {1,0,0}
  slow_axis: {0,-1,0}
  origin: {-212.478,220.002,-190.18}
  distance: 190.18
  pixel to millimeter strategy: SimplePxMmStrategy
Max resolution (at corners): 1.008178
Max resolution (inscribed):  1.204283
Beam:
    wavelength: 0.9795
    sample to source direction : {0,0,1}
    divergence: 0
    sigma divergence: 0
    polarization normal: {0,1,0}
    polarization fraction: 0.999
    flux: 0
    transmission: 1
Beam centre:
    mm: (212.48,220.00)
    px: (1235.34,1279.08)
Scan:
    number of images:   9
    image range:   {1,9}
    oscillation:   {0,0.2}
    exposure time: 0.2
Goniometer:
    Rotation axis:   {1,0,0}
    Fixed rotation:  {1,0,0,0,1,0,0,0,1}
    Setting rotation:{1,0,0,0,1,0,0,0,1}
Crystal:
    Unit cell: 42.272, 42.272, 39.670, 90.000, 89.999, 90.000
    Space group: P 4 2 2
    U matrix:  {{ 0.8336, -0.5360, -0.1335},
                {-0.1798, -0.0348, -0.9831},
                { 0.5223,  0.8435, -0.1254}}
    B matrix:  {{ 0.0237,  0.0000,  0.0000},
                {-0.0000,  0.0237,  0.0000},
                {-0.0000,  0.0000,  0.0252}}
    A = UB:    {{ 0.0197, -0.0127, -0.0034},
                {-0.0043, -0.0008, -0.0248},
                { 0.0124,  0.0200, -0.0032}}
    Mosaicity:  0.157000
""".strip()
    )


def test_dials_show_i04_weak_data(dials_regression):
    path = os.path.join(
        dials_regression,
        "indexing_test_data",
        "i04_weak_data",
        "experiments_import.json",
    )
    result = procrunner.run(
        ["dials.show", path], environment_override={"DIALS_NOBANNER": "1"}
    )
    assert not result.returncode and not result.stderr
    output = result.stdout.decode("latin-1")
    output = [_f for _f in (s.rstrip() for s in output.split("\n")) if _f]
    assert (
        "\n".join(output[6:])
        == """
Detector:
Panel:
  name: Panel
  type: SENSOR_PAD
  identifier:
  pixel_size:{0.172,0.172}
  image_size: {2463,2527}
  trusted_range: {-1,161977}
  thickness: 0
  material:
  mu: 0
  gain: 1
  pedestal: 0
  fast_axis: {1,0,0}
  slow_axis: {0,-1,0}
  origin: {-210.76,205.277,-265.27}
  distance: 265.27
  pixel to millimeter strategy: SimplePxMmStrategy
Max resolution (at corners): 1.161261
Max resolution (inscribed):  1.509475
Beam:
    wavelength: 0.97625
    sample to source direction : {0,0,1}
    divergence: 0
    sigma divergence: 0
    polarization normal: {0,1,0}
    polarization fraction: 0.999
    flux: 0
    transmission: 1
Beam centre:
    mm: (210.76,205.28)
    px: (1225.35,1193.47)
Scan:
    number of images:   540
    image range:   {1,540}
    oscillation:   {82,0.15}
    exposure time: 0.067
Goniometer:
    Rotation axis:   {1,0,0}
    Fixed rotation:  {1,0,0,0,1,0,0,0,1}
    Setting rotation:{1,0,0,0,1,0,0,0,1}
""".strip()
    )


def test_dials_show_centroid_test_data(dials_data):
    result = procrunner.run(
        ["dials.show"]
        + sorted(dials_data("centroid_test_data", pathlib=True).glob("centroid_*.cbf")),
        environment_override={"DIALS_NOBANNER": "1"},
    )
    assert not result.returncode and not result.stderr
    output = result.stdout.decode("latin-1")
    output = [_f for _f in (s.rstrip() for s in output.split("\n")) if _f]
    assert (
        "\n".join(output[7:])
        == """
Detector:
Panel:
  name: Panel
  type: SENSOR_PAD
  identifier:
  pixel_size:{0.172,0.172}
  image_size: {2463,2527}
  trusted_range: {0,495976}
  thickness: 0.32
  material: Si
  mu: 3.96039
  gain: 1
  pedestal: 0
  fast_axis: {1,0,0}
  slow_axis: {0,-1,0}
  origin: {-212.478,220.002,-190.18}
  distance: 190.18
  pixel to millimeter strategy: ParallaxCorrectedPxMmStrategy
    mu: 3.96039
    t0: 0.32
Max resolution (at corners): 1.008375
Max resolution (inscribed):  1.204621
Beam:
    wavelength: 0.9795
    sample to source direction : {0,0,1}
    divergence: 0
    sigma divergence: 0
    polarization normal: {0,1,0}
    polarization fraction: 0.999
    flux: 0
    transmission: 1
Beam centre:
    mm: (212.48,220.00)
    px: (1235.34,1279.08)
Scan:
    number of images:   9
    image range:   {1,9}
    oscillation:   {0,0.2}
    exposure time: 0.2
Goniometer:
    Rotation axis:   {1,0,0}
    Fixed rotation:  {1,0,0,0,1,0,0,0,1}
    Setting rotation:{1,0,0,0,1,0,0,0,1}
""".strip()
    )


def test_dials_show_multi_panel_i23(dials_regression):
    path = os.path.join(
        dials_regression, "image_examples", "DLS_I23", "germ_13KeV_0001.cbf"
    )
    result = procrunner.run(
        ["dials.show", path], environment_override={"DIALS_NOBANNER": "1"}
    )
    assert not result.returncode and not result.stderr
    output = result.stdout.decode("latin-1")
    output = [_f for _f in (s.rstrip() for s in output.split("\n")) if _f]
    assert (
        "\n".join(output[7:27])
        == """
Detector:
Panel:
  name: row-00
  type: SENSOR_PAD
  identifier:
  pixel_size:{0.172,0.172}
  image_size: {2463,195}
  trusted_range: {0,1e+06}
  thickness: 0.32
  material: Si
  mu: 3.663
  gain: 1
  pedestal: 0
  fast_axis: {-1,0,0}
  slow_axis: {0,-0.143467,-0.989655}
  origin: {191.952,-243.628,52.4929}
  distance: 248.638
  pixel to millimeter strategy: ParallaxCorrectedPxMmStrategy
    mu: 3.663
    t0: 0.32
""".strip()
    )

    assert (
        "\n".join(output[-44:])
        == """
Panel:
  name: row-23
  type: SENSOR_PAD
  identifier:
  pixel_size:{0.172,0.172}
  image_size: {2463,195}
  trusted_range: {0,1e+06}
  thickness: 0.32
  material: Si
  mu: 3.663
  gain: 1
  pedestal: 0
  fast_axis: {-1,0,0}
  slow_axis: {-0,-0.0638966,0.997957}
  origin: {191.952,251.939,-0.791525}
  distance: 251.373
  pixel to millimeter strategy: ParallaxCorrectedPxMmStrategy
    mu: 3.663
    t0: 0.32
Max resolution (at corners): 0.624307
Max resolution (inscribed):  0.829324
Beam:
    wavelength: 0.95373
    sample to source direction : {0,0,1}
    divergence: 0
    sigma divergence: 0
    polarization normal: {0,1,0}
    polarization fraction: 0.999
    flux: 0
    transmission: 1
Beam centre:
    mm: panel 12, (191.95,7.22)
    px: panel 12, (1116.00,41.96)
    mm, raw image: (191.95,444.63)
    px, raw image: (1116.00,2585.96)
Scan:
    number of images:   1
    image range:   {1,1}
    oscillation:   {0,0.1}
    exposure time: 0.2
Goniometer:
    Rotation axis:   {-1,0,0}
    Fixed rotation:  {1,0,0,0,1,0,0,0,1}
    Setting rotation:{1,0,0,0,1,0,0,0,1}
""".strip()
    )


def test_dials_show_reflection_table(dials_data):
    """Test the output of dials.show on a reflection_table pickle file"""
    result = procrunner.run(
        [
            "dials.show",
            dials_data("centroid_test_data", pathlib=True) / "integrated.pickle",
        ],
        environment_override={"DIALS_NOBANNER": "1"},
    )
    assert not result.returncode and not result.stderr
    output = result.stdout.decode("latin-1")
    output = [_f for _f in (s.rstrip() for s in output.split("\n")) if _f]

    assert output[4] == "Reflection list contains 2269 reflections"
    headers = ["Column", "min", "max", "mean"]
    for header in headers:
        assert header in output[6]
    row_names = [
        "background.mean",
        "background.sum.value",
        "background.sum.variance",
        "d",
        "dqe",
        "flags",
        "id",
        "imageset_id",
        "intensity.prf.value",
        "intensity.prf.variance",
        "intensity.sum.value",
        "intensity.sum.variance",
        "lp",
        "miller_index",
        "num_pixels.background",
        "num_pixels.background_used",
        "num_pixels.foreground",
        "num_pixels.valid",
        "panel",
        "partial_id",
        "partiality",
        "profile.correlation",
        "profile.rmsd",
        "rlp",
        "s1",
        "shoebox",
        "summed I",
        "N pix",
        "N valid foreground pix",
        "xyzcal.mm",
        "xyzcal.px",
        "xyzobs.mm.value",
        "xyzobs.mm.variance",
        "xyzobs.px.value",
        "xyzobs.px.variance",
        "zeta",
    ]
    for (name, out) in zip(row_names, output[8:-1]):
        assert name in out


def test_dials_show_image_statistics(dials_regression):
    # Run on one multi-panel image
    path = os.path.join(
        dials_regression, "image_examples", "DLS_I23", "germ_13KeV_0001.cbf"
    )
    result = procrunner.run(
        ["dials.show", "image_statistics.show_raw=true", path],
        environment_override={"DIALS_NOBANNER": "1"},
    )
    assert not result.returncode and not result.stderr
    output = result.stdout.decode("latin-1")
    output = [_f for _f in (s.rstrip() for s in output.split("\n")) if _f]
    assert (
        output[-1]
        == "germ_13KeV_0001.cbf: Min: -2.0 Q1: 9.0 Med: 12.0 Q3: 16.0 Max: 1070079.0"
    )


def test_dials_show_image_statistics_with_no_image_data(dials_regression):
    # Example where image data doesn't exist
    path = os.path.join(
        dials_regression, "indexing_test_data", "i04_weak_data", "datablock_orig.json"
    )
    result = procrunner.run(
        ["dials.show", "image_statistics.show_raw=true", path],
        environment_override={"DIALS_NOBANNER": "1"},
    )
    assert result.returncode == 1 and result.stderr


def test_dials_show_on_scaled_data(dials_data):
    """Test that dials.show works on scaled data."""
    location = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
    refl = location / "scaled_30.refl"
    expt = location / "scaled_30.expt"

    result = procrunner.run(["dials.show", refl, expt])
    assert not result.returncode and not result.stderr


def test_model_connectivity(dials_data):
    """Test that dials.show experiments_has_model option."""
    location = dials_data("l_cysteine_dials_output", pathlib=True)
    expts = load.experiment_list(location / "indexed.expt", check_format=False)
    assert (
        model_connectivity(expts)
        == """\
Experiment / Models

Detector:
              0  1
Experiment 0  x  .
Experiment 1  x  .
Experiment 2  x  .
Experiment 3  .  x

Crystal:
              0
Experiment 0  x
Experiment 1  x
Experiment 2  x
Experiment 3  x

Beam:
              0
Experiment 0  x
Experiment 1  x
Experiment 2  x
Experiment 3  x"""
    )


def test_dials_show_shared_models(dials_data, capsys):
    """Test that dials.show experiments_has_model option."""
    location = dials_data("l_cysteine_dials_output", pathlib=True)
    run([str(location / "indexed.expt"), "show_shared_models=True"])
    stdout, stderr = capsys.readouterr()
    assert not stderr
    assert "Experiment / Models" in stdout


def test_dials_show_centroid_test_data_image_zero(dials_data, tmpdir):
    """Integration test: import image 0; show import / show works"""

    im1 = dials_data("centroid_test_data", pathlib=True) / "centroid_0001.cbf"
    im0 = tmpdir.join("centroid_0000.cbf").strpath

    shutil.copyfile(im1, im0)

    result = procrunner.run(("dials.import", im0), working_directory=tmpdir)
    assert not result.returncode and not result.stderr

    result = procrunner.run(("dials.show", "imported.expt"), working_directory=tmpdir)
    assert not result.returncode and not result.stderr



from __future__ import annotations

import math

from scitbx import matrix

"""A Python version of the Rotation::euler_explicit method from Pointless"""


def solve_r3_rotation_for_angles_given_axes(
    R, e1, e2, e3, smaller_phi2_solution=True, return_both_solutions=False, deg=False
):
    """decompose 3*3 rotation matrix R into three rotations about the axes
    e1, e2 and e3

    return angles given axes; return None if no solution
    smaller_phi2_solution = True or False for the two solutions

    Similar to scitbx.math.euler_angles, except the axes here are arbitrary.

    If the angle between e1 & e2 or e2 & e3 are < 90 degrees, then
    there are matrices which have no solution

    The algorithm for decomposing a rotation into three arbitrary rotations was
    described by David Thomas (Thomas, D.J.(1990) Acta Cryst. A46, 321-343)
    and detailed by Gerard Bricogne (Proceedings of the CCP4 Study Weekend 23-24
    January 1987).

    This implementation is a translation of the C++ code Rotation::euler_explicit
    from Pointless and Aimless by Phil Evans.

    The notation used follows Bricogne

    R =  R(e1, phi1) R(e2, phi2) R(e3, phi3)

    where R(e, phi) is the rotation matrix for a rotation by angle phi around
    axis e"""

    assert R.is_r3_rotation_matrix()
    e1 = matrix.col(e1).normalize()
    e2 = matrix.col(e2).normalize()
    e3 = matrix.col(e3).normalize()
    # Fail if e2 & e3 are parallel
    e2xe3 = e2.cross(e3)
    if e2xe3.length_sq() < 1.0e-6:
        return None
    # Make a unit test vector
    u = e2xe3.normalize()
    e1e2 = e1.dot(e2)
    e1e3 = e1.dot(e3)
    e2e3 = e2.dot(e3)
    e1e2e3 = e1.dot(e2.cross(e3))
    Re3 = R * e3
    e1Re3 = e1.dot(Re3)
    # ** Step 1 ** Calculation of phi2  (Bricogne equation (4))
    # e1.(R e3) = (e1.e2)(e2.e3) + {(e1.e3) - (e1.e2)(e2.e3)} cos(phi2)
    #           + (e1.e2 x e3) sin(phi2)
    # The coefficients of cos & sin phi2
    cc = e1e3 - e1e2 * e2e3
    ss = e1e2e3
    # Fail if both are zero (indeterminate)
    if abs(cc) < 1.0e-6 and abs(ss) < 1.0e-6:
        return None
    norm = math.sqrt(cc * cc + ss * ss)
    rhs = (e1Re3 - e1e2 * e2e3) / norm
    # abs(rhs) should not be greater than 1.0, allowing a small tolerance
    if abs(rhs) > 1.000002:
        return None
    if rhs > 1.0:
        rhs = 1.0
    elif rhs < -1.0:
        rhs = -1.0
    cc /= norm
    ss /= norm
    # Solve rhs = cos(phi2) * cc + sin(phi2) * ss
    # using cos(a-b) = cos(a) cos(b) + sin(a) sin(b)
    # where b = phi2
    a = math.atan2(ss, cc)
    amb = math.acos(rhs)
    # Two solutions in range -pi to +pi
    # Note that if e1 == e3, ss = 0, a = 0 & phi2b = -phi2a
    phi2a = a - amb
    if phi2a > math.pi:
        phi2a -= 2.0 * math.pi
    elif phi2a < -math.pi:
        phi2a += 2.0 * math.pi
    phi2b = a + amb
    if phi2b > math.pi:
        phi2b -= 2.0 * math.pi
    elif phi2b < -math.pi:
        phi2b += 2.0 * math.pi
    if return_both_solutions:
        phi2_ = (phi2a, phi2b)
    elif smaller_phi2_solution:
        if abs(phi2a) < abs(phi2b):
            phi2_ = (phi2a,)
        else:
            phi2_ = (phi2b,)
    else:
        if abs(phi2a) > abs(phi2b):
            phi2_ = (phi2a,)
        else:
            phi2_ = (phi2b,)
    solutions = []
    for phi2 in phi2_:
        # ** Step 2 ** Calculation of phi1
        R2 = e2.axis_and_angle_as_r3_rotation_matrix(phi2, deg=False)
        R2inv = R2.transpose()
        v = R2 * e3
        w = Re3
        v1 = v - (v.dot(e1)) * e1
        w1 = w - (w.dot(e1)) * e1
        norm = v1.dot(v1) * w1.dot(w1)
        # If norm = 0, rotations 1 & 3 are around same axis (for this phi2),
        # so any value for phi1 is OK
        if norm > 1.0e-8:
            norm = math.sqrt(norm)
            # cos(phi1) = (v1.w1)/norm
            # sin(phi1) = (v1.w1 x e1)/norm
            phi1 = math.atan2(v1.dot(w1.cross(e1)) / norm, v1.dot(w1) / norm)
            if phi1 > math.pi:
                phi1 -= 2.0 * math.pi
            if phi1 < -math.pi:
                phi1 += 2.0 * math.pi
        else:
            phi1 = 0.0
        # ** Step 3 ** Calculation of phi3
        R1inv = e1.axis_and_angle_as_r3_rotation_matrix(-1.0 * phi1, deg=False)
        R3 = R2inv * R1inv * R
        R3u = R3 * u
        # sin(phi3) = u.R3u x e3
        # cos(phi3) = u.R3u
        phi3 = math.atan2(u.dot(R3u.cross(e3)), u.dot(R3u))
        if deg:
            phi1, phi2, phi3 = tuple([x * 180 / math.pi for x in (phi1, phi2, phi3)])
        solutions.append((phi1, phi2, phi3))

    if return_both_solutions:
        return solutions
    else:
        return solutions[0]



from __future__ import annotations

import pytest

import iotbx.merging_statistics
from cctbx import uctbx
from scitbx.array_family import flex
from scitbx.math import curve_fitting

from dials.util import resolution_analysis


def test_polynomial_fit():
    x = flex.double(range(-50, 50))
    p = (2, 3, 5)
    yo = flex.double(x.size())
    for i in range(len(p)):
        yo += p[i] * flex.pow(x, i)
    yf = resolution_analysis.polynomial_fit(x, yo, degree=2)
    assert yo == pytest.approx(yf)


def test_log_fit():
    x = flex.double(range(0, 100)) * 0.01
    p = (1, 2)
    yo = flex.double(x.size())
    for i in range(len(p)):
        yo += flex.exp(p[i] * flex.pow(x, i))
    yf = resolution_analysis.log_fit(x, yo, degree=2)
    assert yo == pytest.approx(yf, abs=1e-2)


def test_log_inv_fit():
    x = flex.double(range(0, 100)) * 0.01
    p = (1, 2)
    yo = flex.double(x.size())
    for i in range(len(p)):
        yo += 1 / flex.exp(p[i] * flex.pow(x, i))
    yf = resolution_analysis.log_inv_fit(x, yo, degree=2)
    assert yo == pytest.approx(yf, abs=1e-2)


def test_tanh_fit():
    x = flex.double(range(0, 100)) * 0.01
    f = curve_fitting.tanh(0.5, 1.5)
    yo = f(x)
    yf = resolution_analysis.tanh_fit(x, yo)
    assert yo == pytest.approx(yf, abs=1e-5)


@pytest.fixture
def merging_stats(dials_data):
    mtz = str(
        dials_data("x4wide_processed", pathlib=True)
        / "AUTOMATIC_DEFAULT_scaled_unmerged.mtz"
    )
    i_obs, _ = resolution_analysis.miller_array_from_mtz(mtz)
    return iotbx.merging_statistics.dataset_statistics(
        i_obs=i_obs,
        n_bins=20,
        binning_method="counting_sorted",
        use_internal_variance=False,
        eliminate_sys_absent=False,
        assert_is_not_unique_set_under_symmetry=False,
        cc_one_half_significance_level=0.1,
    )


def test_resolution_fit(merging_stats):
    d_star_sq = flex.double(uctbx.d_as_d_star_sq(b.d_min) for b in merging_stats.bins)
    y_obs = flex.double(b.r_merge for b in merging_stats.bins)
    result = resolution_analysis.resolution_fit(
        d_star_sq, y_obs, resolution_analysis.log_inv_fit, 0.6
    )
    assert result.d_min == pytest.approx(1.278, abs=1e-3)
    assert flex.max(flex.abs(result.y_obs - result.y_fit)) < 0.05


def test_resolution_cc_half(merging_stats):
    result = resolution_analysis.resolution_cc_half(merging_stats, limit=0.82)
    assert result.d_min == pytest.approx(1.242, abs=1e-3)
    result = resolution_analysis.resolution_cc_half(
        merging_stats,
        limit=0.82,
        cc_half_method="sigma_tau",
        model=resolution_analysis.polynomial_fit,
    )
    assert result.d_min == pytest.approx(1.233, abs=1e-3)
    assert flex.max(flex.abs(result.y_obs - result.y_fit)) < 0.04
    assert result.critical_values is not None
    assert len(result.critical_values) == len(result.d_star_sq)


def test_resolution_fit_from_merging_stats(merging_stats):
    result = resolution_analysis.resolution_fit_from_merging_stats(
        merging_stats, "i_over_sigma_mean", resolution_analysis.log_fit, limit=1.5
    )
    assert result.d_min == pytest.approx(1.295, abs=1e-3)
    assert flex.max(flex.abs(result.y_obs - result.y_fit)) < 1


def test_resolution_fit_interpolation_error(merging_stats):
    result = resolution_analysis.resolution_fit_from_merging_stats(
        merging_stats, "i_over_sigma_mean", resolution_analysis.log_fit, limit=25
    )
    assert result.d_min is None


def test_plot_result(merging_stats):
    result = resolution_analysis.resolution_cc_half(merging_stats, limit=0.82)
    d = resolution_analysis.plot_result("cc_half", result)
    assert "data" in d
    assert "layout" in d

    result = resolution_analysis.resolution_fit_from_merging_stats(
        merging_stats,
        "unmerged_i_over_sigma_mean",
        resolution_analysis.log_fit,
        limit=0.82,
    )
    d = resolution_analysis.plot_result("isigma", result)
    assert "data" in d
    assert "layout" in d



from __future__ import annotations

import pickle

import procrunner
import pytest

from dxtbx.serialize import load


@pytest.fixture
def model(tmpdir):
    from dials.algorithms.background.gmodel import StaticBackgroundModel
    from dials.array_family import flex

    ysize = 2527
    xsize = 2463
    data = flex.double(flex.grid(ysize, xsize), 1)
    model = StaticBackgroundModel()
    model.add(data)

    model_file = tmpdir.join("model.pickle")
    with model_file.open("wb") as fh:
        pickle.dump(model, fh, pickle.HIGHEST_PROTOCOL)
    return model_file


def test_simple(dials_data, model, tmpdir):
    experiments = dials_data("centroid_test_data", pathlib=True) / "experiments.json"

    reflns_simple = tmpdir.join("simple").join("observations.refl")
    reflns_g_simple = tmpdir.join("gmodel_simple").join("observations.refl")
    reflns_simple.dirpath().ensure(dir=1)
    reflns_g_simple.dirpath().ensure(dir=1)

    # Patched data file. Original had trusted_range from -1, but now this range
    # is defined to start from the minimum trusted value. This test should be
    # updated with new data.
    # https://github.com/dials/dials/issues/2200
    exp = load.experiment_list(experiments)
    panel = exp[0].detector[0]
    max_trusted = panel.get_trusted_range()[1]
    panel.set_trusted_range((0, max_trusted))
    exp.as_json(tmpdir.join("trusted_range_patch.expt"))

    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            "trusted_range_patch.expt",
            "profile.fitting=False",
            "background.algorithm=simple",
            "background.simple.outlier.algorithm=null",
            "output.reflections=" + reflns_simple.strpath,
        ],
        working_directory=tmpdir.strpath,
    )
    assert not result.returncode and not result.stderr
    assert reflns_simple.check()

    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            "trusted_range_patch.expt",
            "profile.fitting=False",
            "background.algorithm=gmodel",
            "background.gmodel.robust.algorithm=False",
            "background.gmodel.model=model.pickle",
            "output.reflections=" + reflns_g_simple.strpath,
        ],
        working_directory=tmpdir.strpath,
    )
    assert not result.returncode and not result.stderr
    assert reflns_g_simple.check()

    from dials.array_family import flex

    reflections1 = flex.reflection_table.from_file(reflns_simple.strpath)
    reflections3 = flex.reflection_table.from_file(reflns_g_simple.strpath)
    assert len(reflections1) == len(reflections3)

    flag = flex.reflection_table.flags.integrated_sum
    integrated1 = reflections1.select(reflections1.get_flags(flag, all=True))
    integrated3 = reflections3.select(reflections3.get_flags(flag, all=True))

    assert len(integrated1) > 0
    assert len(integrated1) == len(integrated3)

    mean_bg1 = integrated1["background.mean"]
    mean_bg3 = integrated3["background.mean"]
    scale3 = integrated3["background.scale"]

    diff1 = flex.abs(mean_bg1 - mean_bg3)
    assert (scale3 > 0).count(False) == 0
    assert (diff1 < 1e-5).count(False) == 0


def test_robust(dials_data, model, tmpdir):
    experiments = dials_data("centroid_test_data", pathlib=True) / "experiments.json"

    reflns_robust = tmpdir.join("robust").join("observations.refl")
    reflns_g_robust = tmpdir.join("gmodel_robust").join("observations.refl")
    reflns_robust.dirpath().ensure(dir=1)
    reflns_g_robust.dirpath().ensure(dir=1)

    # Patched data file. Original had trusted_range from -1, but now this range
    # is defined to start from the minimum trusted value. This test should be
    # updated with new data.
    # https://github.com/dials/dials/issues/2200
    exp = load.experiment_list(experiments)
    panel = exp[0].detector[0]
    max_trusted = panel.get_trusted_range()[1]
    panel.set_trusted_range((0, max_trusted))
    exp.as_json(tmpdir.join("trusted_range_patch.expt"))

    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            "trusted_range_patch.expt",
            "profile.fitting=False",
            "background.algorithm=glm",
            "output.reflections=" + reflns_robust.strpath,
        ],
        working_directory=tmpdir.strpath,
    )
    assert not result.returncode and not result.stderr
    assert reflns_robust.check()

    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            "trusted_range_patch.expt",
            "profile.fitting=False",
            "background.algorithm=gmodel",
            "background.gmodel.robust.algorithm=True",
            "background.gmodel.model=model.pickle",
            "output.reflections=" + reflns_g_robust.strpath,
        ],
        working_directory=tmpdir.strpath,
    )
    assert not result.returncode and not result.stderr
    assert reflns_g_robust.check()

    from dials.array_family import flex

    reflections2 = flex.reflection_table.from_file(reflns_robust.strpath)
    reflections4 = flex.reflection_table.from_file(reflns_g_robust.strpath)
    assert len(reflections2) == len(reflections4)

    flag = flex.reflection_table.flags.integrated_sum
    integrated2 = reflections2.select(reflections2.get_flags(flag, all=True))
    integrated4 = reflections4.select(reflections4.get_flags(flag, all=True))

    assert len(integrated2) > 0
    assert len(integrated2) == len(integrated4)

    scale4 = integrated4["background.scale"]
    assert (scale4 > 0).count(False) == 0



"""
This module defines classes which implement the stages of the scaling algorithm.

These 'scalers' act to initialise and connect various parts of the scaling
algorithm and datastructures such as the Ih_table etc, and
present a united interface to the main scaling algorithm for single, multi
and targeted scaling.

The SingleScaler is defined, for scaling of a single dataset, a MultiScaler is
defined for scaling multiple datasets simultaneously and a TargetScaler is
defined for targeted scaling.
"""

from __future__ import annotations

import copy
import logging
import time
from io import StringIO
from math import ceil

import numpy as np

from dxtbx import flumpy
from libtbx import Auto
from scitbx import sparse

from dials.algorithms.scaling.basis_functions import RefinerCalculator
from dials.algorithms.scaling.combine_intensities import (
    MultiDatasetIntensityCombiner,
    SingleDatasetIntensityCombiner,
)
from dials.algorithms.scaling.error_model.engine import run_error_model_refinement
from dials.algorithms.scaling.Ih_table import IhTable
from dials.algorithms.scaling.outlier_rejection import (
    determine_Esq_outlier_index_arrays,
    determine_outlier_index_arrays,
)
from dials.algorithms.scaling.parameter_handler import ScalingParameterManagerGenerator
from dials.algorithms.scaling.reflection_selection import (
    _select_groups_on_Isigma_cutoff,
    calculate_scaling_subset_ranges,
    calculate_scaling_subset_ranges_with_E2,
    select_connected_reflections_across_datasets,
)
from dials.algorithms.scaling.scaling_library import (
    merging_stats_from_scaled_array,
    scaled_data_as_miller_array,
)
from dials.algorithms.scaling.scaling_refiner import scaling_refinery
from dials.algorithms.scaling.scaling_utilities import (
    BadDatasetForScalingException,
    DialsMergingStatisticsError,
    log_memory_usage,
)
from dials.algorithms.scaling.target_function import ScalingTarget, ScalingTargetFixedIH
from dials.array_family import flex
from dials.util import tabulate
from dials.util.observer import Subject
from dials_scaling_ext import calc_sigmasq as cpp_calc_sigmasq
from dials_scaling_ext import row_multiply

logger = logging.getLogger("dials")

flex.set_random_seed(42)


class ScalerBase(Subject):
    """
    Abstract base class for all scalers (single and multiple).
    """

    def __init__(self, params):
        """Define the properties of a scaler."""
        super().__init__(
            events=[
                "performed_scaling",
                "performed_error_analysis",
                "performed_outlier_rejection",
            ]
        )
        self._params = params
        self._Ih_table = None
        self._global_Ih_table = None
        self._free_Ih_table = None
        self._work_free_stats = []
        self._removed_datasets = []
        self._active_scalers = []

    @property
    def active_scalers(self):
        """A list of scalers that are currently being used in the algorithm."""
        return self._active_scalers

    @property
    def removed_datasets(self):
        """The list of removed datasets."""
        return self._removed_datasets

    @property
    def work_free_stats(self):
        """Holder for work/free set statistics."""
        return self._work_free_stats

    @property
    def Ih_table(self):
        """The Ih_table datastructure for use in minimisation."""
        return self._Ih_table

    @property
    def global_Ih_table(self):
        """
        An Ih_table datastructure containing all suitable reflections.

        This includes reflections across all datasets being minimised, and there
        should only be one instance, maintained by the highest level scaler, e.g.
        a multiscaler in a multi-dataset case.
        """
        return self._global_Ih_table

    @property
    def params(self):
        """The params phil scope."""
        return self._params

    ### Interface for scaling refiner

    def update_for_minimisation(self, apm, block_id):
        """Update the scale factors and Ih for the next minimisation iteration."""
        raise NotImplementedError()

    def get_blocks_for_minimisation(self):
        """Return the blocks to iterate over during refinement."""
        if self.Ih_table.free_Ih_table:
            return self.Ih_table.blocked_data_list[:-1]
        return self.Ih_table.blocked_data_list

    ## Interface for algorithms using the scaler

    def round_of_outlier_rejection(self):
        """Perform a round of outlier rejection on the reflections."""
        raise NotImplementedError()

    def expand_scales_to_all_reflections(self, caller=None, calc_cov=False):
        """Expand scales from a subset to all reflections."""
        raise NotImplementedError()

    def combine_intensities(self):
        """Combine prf and sum intensities to give optimal intensities."""
        pass

    def make_ready_for_scaling(self, outlier=True):
        """Make the scaler in a prepared state for scaling."""
        raise NotImplementedError()

    def determine_shared_model_components(self):
        return None

    @Subject.notify_event(event="performed_scaling")
    def perform_scaling(self, engine=None, max_iterations=None, tolerance=None):
        """Minimise the scaling model."""
        self._perform_scaling(
            target_type=ScalingTarget,
            engine=engine,
            max_iterations=max_iterations,
            tolerance=tolerance,
        )

    def _perform_scaling(
        self, target_type, engine=None, max_iterations=None, tolerance=None
    ):
        target = target_type()
        # find if any components to share
        shared = self.determine_shared_model_components()
        pmg = ScalingParameterManagerGenerator(
            self.active_scalers,
            target,
            self.params.scaling_refinery.refinement_order,
            shared=shared,
        )
        for apm in pmg.parameter_managers():
            if not engine:
                engine = self.params.scaling_refinery.engine
            if not max_iterations:
                max_iterations = self.params.scaling_refinery.max_iterations
            st = time.time()
            refinery = scaling_refinery(
                engine=engine,
                scaler=self,
                target=target,
                prediction_parameterisation=apm,
                max_iterations=max_iterations,
            )
            if tolerance:
                refinery.set_tolerance(tolerance)
            try:
                refinery.run()
            except RuntimeError as e:
                logger.error(e, exc_info=True)
            logger.info("Time taken for refinement %.2f", (time.time() - st))
            refinery.print_step_table()
            self._update_after_minimisation(apm)
            logger.info("\n" + "=" * 80 + "\n")

    def clear_Ih_table(self):
        """Delete the data from the current Ih_table."""
        self._Ih_table = []

    def fix_initial_parameter(self):
        return False

    def prepare_reflection_tables_for_output(self):
        """Finish adjust reflection table data at the end of the algorithm."""
        # First adjust variances
        for scaler in self.active_scalers:
            if scaler._experiment.scaling_model.error_model:
                scaler.reflection_table["variance"] = flumpy.from_numpy(
                    scaler._experiment.scaling_model.error_model.update_variances(
                        scaler.reflection_table["variance"],
                        scaler.reflection_table["intensity"],
                    )
                )
            # now increase the errors slightly to take into account the uncertainty in the
            # inverse scale factors
            if (
                scaler.var_cov_matrix.non_zeroes > 0
            ):  # parameters errors have been determined
                fractional_error = (
                    flex.sqrt(scaler.reflection_table["inverse_scale_factor_variance"])
                    / scaler.reflection_table["inverse_scale_factor"]
                )
                variance_scaling = (
                    flex.double(scaler.reflection_table.size(), 1.0) + fractional_error
                )
                scaler.reflection_table["variance"] *= variance_scaling
        self._set_outliers()
        self._clean_reflection_tables()
        msg = """
The reflection table variances have been adjusted to account for the
uncertainty in the scaling model""" + (
            "s for all datasets" if len(self.active_scalers) > 1 else ""
        )
        logger.info(msg)
        if self._free_Ih_table:
            # calc merging stats and log.
            free_miller_array = scaled_data_as_miller_array(
                [self.get_free_set_reflections()],
                [self.active_scalers[0].experiment],
                anomalous_flag=False,
            )
            work_miller_array = scaled_data_as_miller_array(
                [self.get_work_set_reflections()],
                [self.active_scalers[0].experiment],
                anomalous_flag=False,
            )
            free, _ = merging_stats_from_scaled_array(
                free_miller_array,
                self.params.output.merging.nbins,
                self.params.output.use_internal_variance,
                anomalous=False,
            )
            s = StringIO()
            s.write("\nFree set statistics\n")
            free.show(out=s)
            work, _ = merging_stats_from_scaled_array(
                work_miller_array,
                self.params.output.merging.nbins,
                self.params.output.use_internal_variance,
                anomalous=False,
            )
            s.write("\nWork set statistics\n")
            work.show(out=s)
            logger.debug(s.getvalue())
            self._work_free_stats = [
                work.overall.r_meas,
                free.overall.r_meas,
                free.overall.r_meas - work.overall.r_meas,
                work.overall.cc_one_half,
                free.overall.cc_one_half,
                work.overall.cc_one_half - free.overall.cc_one_half,
            ]

    # Internal general scaler methods

    def _set_outliers(self):
        """Set the scaling outliers in the individual reflection tables."""
        for scaler in self.active_scalers:
            suitable_isel = scaler.suitable_refl_for_scaling_sel.iselection()
            outlier_isel = suitable_isel.select(scaler.outliers)
            n = scaler.reflection_table.size()
            outliers_mask = flex.bool(n, False)
            outliers_mask.set_selected(outlier_isel, True)
            scaler.reflection_table.unset_flags(
                flex.bool(n, True), scaler.reflection_table.flags.outlier_in_scaling
            )
            scaler.reflection_table.set_flags(
                outliers_mask, scaler.reflection_table.flags.outlier_in_scaling
            )

    def _clean_reflection_tables(self):
        """Remove unnecessary columns added to reflection tables."""
        for scaler in self.active_scalers:
            scaler.clean_reflection_table()

    def _update_after_minimisation(self, parameter_manager):
        if parameter_manager.apm_list[0].var_cov_matrix:
            for i, scaler in enumerate(self.active_scalers):
                scaler.update_var_cov(parameter_manager.apm_list[i])


class SingleScaler(ScalerBase):
    """Definition of a scaler for a single dataset."""

    id_ = "single"

    def __init__(self, params, experiment, reflection_table, for_multi=False):
        """
        Initialise a single-dataset scaler.

        The reflection table needs the columns 'inverse_scale_factor', 'Esq',
        'intensity', 'variance', 'id', which are guaranteed if the scaler is
        created using the SingleScalerFactory.
        """
        assert all(
            i in reflection_table
            for i in ["inverse_scale_factor", "intensity", "variance", "id"]
        )
        super().__init__(params)
        self._experiment = experiment
        n_model_params = sum(val.n_params for val in self.components.values())
        self._var_cov_matrix = sparse.matrix(n_model_params, n_model_params)
        self._initial_keys = list(reflection_table.keys())
        self._reflection_table = reflection_table
        self._Ih_table = None  # stores data for reflections used for minimisation
        self.suitable_refl_for_scaling_sel = self._get_suitable_for_scaling_sel(
            self._reflection_table
        )
        self.n_suitable_refl = self.suitable_refl_for_scaling_sel.count(True)
        if self._experiment.scaling_model.is_scaled:
            outliers = self._reflection_table.get_flags(
                self._reflection_table.flags.outlier_in_scaling
            )
            self.outliers = outliers.select(self.suitable_refl_for_scaling_sel)
        else:
            self.outliers = flex.bool(self.n_suitable_refl, False)
        self.scaling_subset_sel = (
            None  # A selection of len n_suitable_refl of scaling subset selection
        )
        self.scaling_selection = None  # As above, but with outliers deselected also
        self.free_set_selection = flex.bool(self.n_suitable_refl, False)
        self._free_Ih_table = None  # An array of len n_suitable_refl
        self._configure_model_and_datastructures(for_multi=for_multi)
        if self.params.weighting.error_model.error_model:
            # reload current error model parameters, or create new null
            self.experiment.scaling_model.load_error_model(
                self.params.weighting.error_model
            )
            self._update_error_model(self.experiment.scaling_model.error_model)
        if "Imid" in self.experiment.scaling_model.configdict:
            self._combine_intensities(self.experiment.scaling_model.configdict["Imid"])
        if not self._experiment.scaling_model.is_scaled:
            self.round_of_outlier_rejection()
        if not for_multi:
            self._select_reflections_for_scaling()
            self._create_Ih_table()
            self._update_model_data()
        else:
            self._global_Ih_table = None
            self.scaling_selection = ~self.outliers
        logger.info(
            "Completed preprocessing and initialisation for this dataset.\n"
            "\n" + "=" * 80 + "\n"
        )
        log_memory_usage()

    @Subject.notify_event(event="performed_error_analysis")
    def perform_error_optimisation(self, update_Ih=True):
        """Perform an optimisation of the sigma values."""
        # error model should be determined using anomalous groups
        Ih_table, _ = self._create_global_Ih_table(anomalous=True, remove_outliers=True)
        try:
            model = run_error_model_refinement(
                self._experiment.scaling_model.error_model, Ih_table
            )
        except (ValueError, RuntimeError) as e:
            logger.info(e)
            logger.debug(e, exc_info=True)
        else:
            self._update_error_model(model, update_Ih=update_Ih)

    def _update_error_model(self, error_model, update_Ih=True):
        """Update the error model in Ih table."""
        if update_Ih and error_model:
            self.global_Ih_table.update_weights(error_model)
            if self._free_Ih_table:
                self._free_Ih_table.update_weights(error_model)
        self._experiment.scaling_model.set_error_model(error_model)

    def get_valid_reflections(self):
        """All reflections not bad for scaling or user excluded."""
        return self.reflection_table.select(self.suitable_refl_for_scaling_sel)

    def get_work_set_reflections(self):
        valid = self.get_valid_reflections()
        return valid.select(~self.free_set_selection)

    def get_free_set_reflections(self):
        """Get all reflections in the free set if it exists."""
        valid = self.get_valid_reflections()
        return valid.select(self.free_set_selection)

    def get_reflections_for_model_minimisation(self):
        valid = self.get_valid_reflections()
        return valid.select(self.scaling_selection)

    @property
    def active_scalers(self):
        return [self]

    @property
    def experiment(self):
        """The experiment object associated with the dataset."""
        return self._experiment

    @property
    def reflection_table(self):
        """The reflection table of the dataset."""
        return self._reflection_table

    @reflection_table.setter
    def reflection_table(self, new_table):
        """Set the reflection table of the dataset."""
        self._reflection_table = new_table

    def fix_initial_parameter(self):
        fixed = self.experiment.scaling_model.fix_initial_parameter(self.params)
        return fixed

    @staticmethod
    def _get_suitable_for_scaling_sel(reflections):
        """Extract suitable reflections for scaling from the reflection table."""
        user_excl = reflections.get_flags(reflections.flags.user_excluded_in_scaling)
        excl_for_scale = reflections.get_flags(reflections.flags.excluded_for_scaling)
        suitable_refl_for_scaling_sel = ~(user_excl | excl_for_scale)
        return suitable_refl_for_scaling_sel

    @property
    def components(self):
        """Shortcut to the scaling model components."""
        return self.experiment.scaling_model.components

    @property
    def fixed_components(self):
        """Shortcut to the scaling model fixed components."""
        return self.experiment.scaling_model.fixed_components

    @property
    def consecutive_refinement_order(self):
        """Link to consecutive refinement order for parameter manager."""
        return self.experiment.scaling_model.consecutive_refinement_order

    @property
    def var_cov_matrix(self):
        """The variance covariance matrix for the parameters."""
        return self._var_cov_matrix

    def update_var_cov(self, apm):
        """
        Update the full parameter variance covariance matrix after a refinement.

        If all parameters have been refined, then the full var_cov matrix can be set.
        Else one must select subblocks for pairs of parameters and assign these into
        the full var_cov matrix, taking care to out these in the correct position.
        This is applicable if only some parameters have been refined in this cycle.
        """
        var_cov_list = apm.var_cov_matrix  # values are passed as a list from refinery
        if int(var_cov_list.size() ** 0.5) == self.var_cov_matrix.n_rows:
            self._var_cov_matrix.assign_block(
                var_cov_list.matrix_copy_block(
                    0, 0, apm.n_active_params, apm.n_active_params
                ),
                0,
                0,
            )
        else:  # need to set part of the var_cov matrix e.g. if only refined some params
            # first work out the order in self._var_cov_matrix
            cumul_pos_dict = {}
            n_cumul_params = 0
            for name, component in self.components.items():
                cumul_pos_dict[name] = n_cumul_params
                n_cumul_params += component.n_params
            # now get a var_cov_matrix subblock for pairs of parameters
            for name in apm.components_list:
                for name2 in apm.components_list:
                    n_rows = apm.components[name]["n_params"]
                    n_cols = apm.components[name2]["n_params"]
                    start_row = apm.components[name]["start_idx"]
                    start_col = apm.components[name2]["start_idx"]
                    sub = var_cov_list.matrix_copy_block(
                        start_row, start_col, n_rows, n_cols
                    )
                    # now set this block into correct location in overall var_cov
                    self._var_cov_matrix.assign_block(
                        sub, cumul_pos_dict[name], cumul_pos_dict[name2]
                    )

    def update_for_minimisation(self, apm, block_id):
        """Update the scale factors and Ih for the next minimisation iteration."""
        apm_i = apm.apm_list[0]
        scales_i, derivs_i = RefinerCalculator.calculate_scales_and_derivatives(
            apm_i, block_id
        )
        self.Ih_table.set_derivatives(derivs_i, block_id)
        self.Ih_table.set_inverse_scale_factors(flumpy.to_numpy(scales_i), block_id)
        self.Ih_table.calc_Ih(block_id)

    def combine_intensities(self):
        """Combine prf and sum intensities to give optimal intensities."""
        self._combine_intensities()

    def _combine_intensities(self, use_Imid=None):
        try:
            if use_Imid is not None:
                logger.info(
                    "Using previously determined optimal intensity choice: %s\n",
                    {
                        use_Imid: str(round(use_Imid, 4)),
                        0: "profile intensities",
                        1: "summation intensities",
                    }[use_Imid],
                )
            else:
                logger.info("Performing profile/summation intensity optimisation.")
            combiner = SingleDatasetIntensityCombiner(self, use_Imid)
        except DialsMergingStatisticsError as e:
            logger.info("Intensity combination failed with the error %s", e)
        else:
            intensity, variance = combiner.calculate_suitable_combined_intensities()
            # update data in reflection table
            isel = self.suitable_refl_for_scaling_sel.iselection()
            self._reflection_table["intensity"].set_selected(isel, intensity)
            self._reflection_table["variance"].set_selected(isel, variance)
            if self.global_Ih_table:
                self.global_Ih_table.update_data_in_blocks(
                    intensity, 0, column="intensity"
                )
                self.global_Ih_table.update_data_in_blocks(
                    variance, 0, column="variance"
                )
                self.global_Ih_table.update_weights(
                    self._experiment.scaling_model.error_model
                )
                self.global_Ih_table.calc_Ih()
            if self._free_Ih_table:
                self._free_Ih_table.update_data_in_blocks(
                    intensity, 0, column="intensity"
                )
                self._free_Ih_table.update_data_in_blocks(
                    variance, 0, column="variance"
                )
                self._free_Ih_table.update_weights(
                    self._experiment.scaling_model.error_model
                )
                self._free_Ih_table.calc_Ih()
            self.experiment.scaling_model.record_intensity_combination_Imid(
                combiner.max_key
            )

    def expand_scales_to_all_reflections(self, caller=None, calc_cov=False):
        """
        Calculate scale factors for all suitable reflections.

        Use the current model to calculate scale factors for all suitable
        reflections, and set these in the reflection table. If caller=None,
        the global_Ih_table is updated. If calc_cov, an error estimate on the
        inverse scales is calculated.
        """
        self._reflection_table["inverse_scale_factor_variance"] = flex.double(
            self.reflection_table.size(), 0.0
        )
        n_blocks = self.params.scaling_options.nproc
        n_start = 0
        all_scales = flex.double([])
        all_invsfvars = flex.double([])
        n_param_tot = sum(c.n_params for c in self.components.values())
        for i in range(1, n_blocks + 1):  # do calc in blocks for speed/memory
            n_end = int(i * self.n_suitable_refl / n_blocks)
            block_isel = flex.size_t(range(n_start, n_end))
            n_start = n_end
            scales = flex.double(block_isel.size(), 1.0)
            scales_list = []
            derivs_list = []
            jacobian = sparse.matrix(block_isel.size(), n_param_tot)
            for component in self.components.values():
                component.update_reflection_data(block_selections=[block_isel])
                comp_scales, d = component.calculate_scales_and_derivatives(block_id=0)
                scales_list.append(comp_scales)
                if calc_cov:
                    derivs_list.append(d)
                scales *= comp_scales
            all_scales.extend(scales)
            if calc_cov and self.var_cov_matrix.non_zeroes > 0:
                n_cumulative_param = 0
                for j, component in enumerate(self.components):
                    d_block = derivs_list[j]
                    n_param = self.components[component].n_params
                    for k, component_2 in enumerate(self.components):
                        if component_2 != component:
                            d_block = row_multiply(d_block, scales_list[k])
                    jacobian.assign_block(d_block, 0, n_cumulative_param)
                    n_cumulative_param += n_param
                all_invsfvars.extend(
                    cpp_calc_sigmasq(jacobian.transpose(), self._var_cov_matrix)
                )
        scaled_isel = self.suitable_refl_for_scaling_sel.iselection()
        self.reflection_table["inverse_scale_factor"].set_selected(
            scaled_isel, all_scales
        )
        if calc_cov and self.var_cov_matrix.non_zeroes > 0:
            self.reflection_table["inverse_scale_factor_variance"].set_selected(
                scaled_isel, all_invsfvars
            )
        if caller is None:
            self.global_Ih_table.update_data_in_blocks(
                self.reflection_table["inverse_scale_factor"].select(
                    self.suitable_refl_for_scaling_sel
                ),
                dataset_id=0,
                column="inverse_scale_factor",
            )
            self.global_Ih_table.calc_Ih()
            if self._free_Ih_table:
                self._free_Ih_table.update_data_in_blocks(
                    self.reflection_table["inverse_scale_factor"].select(
                        self.suitable_refl_for_scaling_sel
                    ),
                    dataset_id=0,
                    column="inverse_scale_factor",
                )
                self._free_Ih_table.calc_Ih()
            logger.info(
                "Scale factors determined during minimisation have now been\n"
                "applied to all reflections for dataset %s.\n",
                self.reflection_table["id"][0],
            )

    def _select_reflections_for_scaling(self):
        """Select a subset of reflections to use in minimisation."""
        # For single dataset, auto means random
        if self.params.reflection_selection.method in (
            None,
            Auto,
            "auto",
            "random",
            "quasi_random",
        ):
            # do the random reflection selection.
            block = self.global_Ih_table.Ih_table_blocks[0]
            suitable_table = self.get_valid_reflections()
            presel = flumpy.to_numpy(
                calculate_scaling_subset_ranges(
                    suitable_table, self.params, print_summary=True
                )
            )
            preselection = presel[block.Ih_table["loc_indices"].to_numpy()]
            presel_block = block.select(preselection)
            # then select random groups
            n_h_over_1 = presel_block.calc_nh() > 1
            if n_h_over_1.any():
                presel_block = presel_block.select(n_h_over_1)
                min_groups = self.params.reflection_selection.random.min_groups
                avg_multi = np.mean(presel_block.group_multiplicities())
                min_refl = self.params.reflection_selection.random.min_reflections
                n_groups_in_table = presel_block.n_groups
                n_groups_to_sel = max(int(ceil(min_refl / avg_multi)), min_groups)

                logger.debug(
                    "Average multiplicity that reflection selection is sampling from: %s",
                    avg_multi,
                )
                if n_groups_to_sel < n_groups_in_table:
                    isel = flumpy.to_numpy(
                        flex.random_selection(n_groups_in_table, n_groups_to_sel)
                    )
                    loc_indices = (
                        presel_block.select_on_groups(isel)
                        .Ih_table["loc_indices"]
                        .to_numpy()
                    )
                else:
                    loc_indices = presel_block.Ih_table["loc_indices"].to_numpy()
                    n_groups_to_sel = n_groups_in_table
                self.scaling_selection = np.full(
                    self.n_suitable_refl, False, dtype=bool
                )
                # flex.bool(self.n_suitable_refl, False)
                self.scaling_selection[loc_indices] = True
                self.scaling_selection = flumpy.from_numpy(self.scaling_selection)
                logger.info(
                    """Randomly selected %s/%s groups (m>1) to use for scaling model
minimisation (%s reflections)""",
                    n_groups_to_sel,
                    n_groups_in_table,
                    loc_indices.size,
                )
            else:
                logger.warning(
                    """No groups left with multiplicity >1 after prefiltering,
attempting to use all reflections for minimisation."""
                )
                self.params.reflection_selection.method = "use_all"

        if self.params.reflection_selection.method == "intensity_ranges":
            overall_scaling_selection = calculate_scaling_subset_ranges_with_E2(
                self.reflection_table, self.params
            )
            self.scaling_selection = overall_scaling_selection.select(
                self.suitable_refl_for_scaling_sel
            )
            if self._free_Ih_table:
                self.scaling_selection.set_selected(self.free_set_selection, False)
        if self.params.reflection_selection.method == "use_all":
            if self._free_Ih_table:
                self.scaling_selection = ~self.free_set_selection
            else:
                self.scaling_selection = flex.bool(self.n_suitable_refl, True)
        self.scaling_subset_sel = copy.deepcopy(self.scaling_selection)
        self.scaling_selection &= ~self.outliers  # now apply outliers

    def _update_model_data(self):
        """Use the data in the Ih_table to update the model data."""
        assert self.Ih_table is not None
        block_selections = self.Ih_table.get_block_selections_for_dataset(dataset=0)
        block_selections = [flumpy.from_numpy(b) for b in block_selections]
        for component in self.components.values():
            component.update_reflection_data(block_selections=block_selections)

    def _create_Ih_table(self):
        """Create an Ih_table from the reflection table using the scaling selection."""
        self._Ih_table = IhTable(
            [self.get_reflections_for_model_minimisation()],
            self.experiment.crystal.get_space_group(),
            indices_lists=[self.scaling_selection.iselection()],
            nblocks=self.params.scaling_options.nproc,
            anomalous=self.params.anomalous,
        )
        if self._experiment.scaling_model.error_model:
            # update with the error model to add the correct weights
            self._Ih_table.update_weights(self._experiment.scaling_model.error_model)

    def _create_global_Ih_table(
        self, free_set_percentage=0, anomalous=False, remove_outliers=False
    ):
        sel_reflections = self.get_valid_reflections()
        if remove_outliers:
            sel_reflections = sel_reflections.select(~self.outliers)
        free_Ih_table = None
        global_Ih_table = IhTable(
            [sel_reflections],
            self.experiment.crystal.get_space_group(),
            nblocks=1,
            free_set_percentage=free_set_percentage,
            free_set_offset=self.params.scaling_options.free_set_offset,
            anomalous=anomalous,
        )
        if free_set_percentage:
            loc_indices = (
                global_Ih_table.blocked_data_list[-1].Ih_table["loc_indices"].to_numpy()
            )
            self.free_set_selection = flex.bool(self.n_suitable_refl, False)
            self.free_set_selection.set_selected(flumpy.from_numpy(loc_indices), True)
            global_Ih_table = IhTable(
                [sel_reflections.select(~self.free_set_selection)],
                self.experiment.crystal.get_space_group(),
                [(~self.free_set_selection).iselection()],
                nblocks=1,
                anomalous=anomalous,
            )
            free_Ih_table = IhTable(
                [sel_reflections.select(self.free_set_selection)],
                self.experiment.crystal.get_space_group(),
                [self.free_set_selection.iselection()],
                nblocks=1,
                anomalous=anomalous,
            )
        return global_Ih_table, free_Ih_table

    def _configure_model_and_datastructures(self, for_multi=False):
        """
        Store the relevant data in the scaling model components.

        This takes the columns from the 'suitable' part of the reflection table (
        which will include outliers). Then a global_Ih_table is created, which can
        be used for outlier rejection. When calculations are done with the model
        components, the correct reflections should first be selected out of the
        stored data.
        """
        sel_reflections = self.get_valid_reflections()
        self.experiment.scaling_model.configure_components(
            sel_reflections, self.experiment, self.params
        )
        free_set_percentage = 0
        if self.params.scaling_options.use_free_set and not for_multi:
            free_set_percentage = self.params.scaling_options.free_set_percentage
        self._global_Ih_table, self._free_Ih_table = self._create_global_Ih_table(
            free_set_percentage=free_set_percentage, anomalous=self.params.anomalous
        )
        rows = [[key, str(val.n_params)] for key, val in self.components.items()]
        logger.info("The following corrections will be applied to this dataset: \n")
        logger.info(tabulate(rows, ["correction", "n_parameters"]))

    @Subject.notify_event(event="performed_outlier_rejection")
    def round_of_outlier_rejection(self):
        """Perform a round of outlier rejection, set a new outliers array."""
        assert self.global_Ih_table is not None
        if self.params.scaling_options.outlier_rejection:
            outlier_indices = determine_outlier_index_arrays(
                self.global_Ih_table,
                self.params.scaling_options.outlier_rejection,
                self.params.scaling_options.outlier_zmax,
            )[0]
            outlier_indices = flumpy.from_numpy(outlier_indices)
            self.outliers = flex.bool(self.n_suitable_refl, False)
            self.outliers.set_selected(outlier_indices, True)
            if self.params.scaling_options.emax:
                Esq_outlier_indices = determine_Esq_outlier_index_arrays(
                    self.global_Ih_table,
                    self.experiment,
                    self.params.scaling_options.emax,
                )[0]
                self.outliers.set_selected(flumpy.from_numpy(Esq_outlier_indices), True)
            if self._free_Ih_table:
                free_outlier_indices = determine_outlier_index_arrays(
                    self._free_Ih_table,
                    self.params.scaling_options.outlier_rejection,
                    self.params.scaling_options.outlier_zmax,
                )[0]
                self.outliers.set_selected(
                    flumpy.from_numpy(free_outlier_indices), True
                )
                if self.params.scaling_options.emax:
                    free_Esq_outlier_indices = determine_Esq_outlier_index_arrays(
                        self._free_Ih_table,
                        self.experiment,
                        self.params.scaling_options.emax,
                    )[0]
                    self.outliers.set_selected(
                        flumpy.from_numpy(free_Esq_outlier_indices), True
                    )

    def make_ready_for_scaling(self, outlier=True):
        """
        Prepare the datastructures for a round of scaling.

        Update the scaling selection, create a new Ih_table and update the model
        data ready for minimisation.
        """
        if outlier:
            self.scaling_selection = self.scaling_subset_sel & ~self.outliers
        else:
            self.scaling_selection = copy.deepcopy(self.scaling_subset_sel)
        self._create_Ih_table()
        self._update_model_data()

    def clean_reflection_table(self):
        """Remove additional added columns that are not required for output."""
        self._initial_keys.append("inverse_scale_factor")
        self._initial_keys.append("inverse_scale_factor_variance")
        self._initial_keys.append("Ih_values")
        self._initial_keys.append("intensity.scale.value")
        self._initial_keys.append("intensity.scale.variance")
        self.reflection_table["intensity.scale.value"] = self.reflection_table[
            "intensity"
        ]
        self.reflection_table["intensity.scale.variance"] = self.reflection_table[
            "variance"
        ]
        if "Esq" in self.reflection_table:
            del self.reflection_table["Esq"]
        for key in self.reflection_table.keys():
            if key not in self._initial_keys:
                del self._reflection_table[key]
        bad = self._reflection_table.get_flags(
            self._reflection_table.flags.bad_for_scaling, all=False
        )
        self._reflection_table.set_flags(~bad, self.reflection_table.flags.scaled)


class MultiScalerBase(ScalerBase):
    """Base class for scalers handling multiple datasets."""

    def __init__(self, single_scalers):
        """Initialise from a list of single scalers."""
        super().__init__(single_scalers[0].params)
        self.single_scalers = single_scalers

    def remove_datasets(self, scalers, n_list):
        """
        Delete a scaler from the dataset.

        Code in this module does not necessarily have access to all references of
        experiments and reflections, so log the position in the list so that they
        can be deleted later. Scaling algorithm code should only depends on the
        scalers.
        """
        initial_number = len(scalers)
        for n in n_list[::-1]:
            self._removed_datasets.append(scalers[n].experiment.identifier)
            del scalers[n]
        # if 0 in n_list:
        #    self._experiment = scalers[0].experiments
        assert len(scalers) == initial_number - len(n_list)
        logger.info("Removed datasets: %s", n_list)
        if not self.active_scalers:
            raise ValueError("No datasets remain after filtering out bad datasets.")

    def get_free_set_reflections(self):
        """Get all reflections in the free set if it exists."""
        if self._free_Ih_table:
            refls = flex.reflection_table()
            for scaler in self.active_scalers:
                refls.extend(scaler.get_free_set_reflections())
            return refls
        return None

    def get_work_set_reflections(self):
        """Get all reflections in the free set if it exists."""
        if self._free_Ih_table:
            refls = flex.reflection_table()
            for scaler in self.active_scalers:
                refls.extend(scaler.get_work_set_reflections())
            return refls
        return None

    def expand_scales_to_all_reflections(self, caller=None, calc_cov=False):
        """
        Calculate scale factors for all suitable reflections in the datasets.

        After the scale factors are updated, the global_Ih_table is updated also.
        """
        if calc_cov:
            logger.info("Calculating error estimates of inverse scale factors. \n")
        for i, scaler in enumerate(self.active_scalers):
            scaler.expand_scales_to_all_reflections(caller=self, calc_cov=calc_cov)
            # now update global Ih table
            self.global_Ih_table.update_data_in_blocks(
                scaler.reflection_table["inverse_scale_factor"].select(
                    scaler.suitable_refl_for_scaling_sel
                ),
                dataset_id=i,
                column="inverse_scale_factor",
            )
            if self._free_Ih_table:
                self._free_Ih_table.update_data_in_blocks(
                    scaler.reflection_table["inverse_scale_factor"].select(
                        scaler.suitable_refl_for_scaling_sel
                    ),
                    dataset_id=i,
                    column="inverse_scale_factor",
                )

        self.global_Ih_table.calc_Ih()
        if self._free_Ih_table:
            self._free_Ih_table.calc_Ih()
        logger.info(
            "Scale factors determined during minimisation have now been\n"
            "applied to all datasets.\n"
        )

    def update_for_minimisation(self, apm, block_id):
        """Update the scale factors and Ih for the next iteration of minimisation."""
        self._update_for_minimisation(apm, block_id, calc_Ih=True)

    def _update_for_minimisation(self, apm, block_id, calc_Ih=True):
        scales = flex.double([])
        derivs = []
        for apm_i in apm.apm_list:
            scales_i, derivs_i = RefinerCalculator.calculate_scales_and_derivatives(
                apm_i, block_id
            )
            scales.extend(scales_i)
            derivs.append(derivs_i)
        deriv_matrix = sparse.matrix(scales.size(), apm.n_active_params)
        start_row_no = 0
        for j, deriv in enumerate(derivs):
            deriv_matrix.assign_block(deriv, start_row_no, apm.apm_data[j]["start_idx"])
            start_row_no += deriv.n_rows
        self.Ih_table.set_inverse_scale_factors(flumpy.to_numpy(scales), block_id)
        self.Ih_table.set_derivatives(deriv_matrix, block_id)
        if calc_Ih:
            self.Ih_table.calc_Ih(block_id)
        # The parallelisation below would work if sparse matrices were
        # pickleable (I think!) - with more benefit for larger number of datasets.'''
        """def task_wrapper(block):
            s, d = RefinerCalculator.calculate_scales_and_derivatives(block)
            return s, d
        blocks = apm.apm_list
        task_results = easy_mp.parallel_map(func=task_wrapper, iterable=blocks,
            processes=n_datasets, method="multiprocessing",
            preserve_exception_message=True
        )
        scales_list, derivs_list = zip(*task_results)"""

    def _update_model_data(self):
        for i, scaler in enumerate(self.active_scalers):
            block_selections = self.Ih_table.get_block_selections_for_dataset(i)
            these_block_selections = [flumpy.from_numpy(b) for b in block_selections]
            for component in scaler.components.values():
                component.update_reflection_data(
                    block_selections=these_block_selections
                )

    def _create_global_Ih_table(self, anomalous=False, remove_outliers=False):
        if remove_outliers:
            tables = [
                s.get_valid_reflections().select(~s.outliers)
                for s in self.active_scalers
            ]
        else:
            tables = [s.get_valid_reflections() for s in self.active_scalers]
        free_set_percentage = 0.0
        free_Ih_table = None
        if self.params.scaling_options.use_free_set:
            free_set_percentage = self.params.scaling_options.free_set_percentage
        space_group = self.active_scalers[0].experiment.crystal.get_space_group()
        global_Ih_table = IhTable(
            tables,
            space_group,
            nblocks=1,
            additional_cols=["partiality"],
            free_set_percentage=free_set_percentage,
            free_set_offset=self.params.scaling_options.free_set_offset,
            anomalous=anomalous,
        )
        if free_set_percentage:
            # need to set free_set_selection in individual scalers
            tables = []
            free_tables = []
            indices_list = []
            free_indices_list = []
            for i, scaler in enumerate(self.active_scalers):
                sel = (
                    global_Ih_table.Ih_table_blocks[-1]
                    .Ih_table["dataset_id"]
                    .to_numpy()
                    == i
                )
                indiv_Ih_block = global_Ih_table.Ih_table_blocks[-1].select(sel)
                loc_indices = indiv_Ih_block.Ih_table["loc_indices"].to_numpy()
                scaler.free_set_selection = flex.bool(scaler.n_suitable_refl, False)
                scaler.free_set_selection.set_selected(
                    flumpy.from_numpy(loc_indices), True
                )
                tables.append(scaler.get_work_set_reflections())
                free_tables.append(scaler.get_free_set_reflections())
                free_indices_list.append(scaler.free_set_selection.iselection())
                indices_list.append((~scaler.free_set_selection).iselection())
            global_Ih_table = IhTable(
                tables,
                space_group,
                indices_list,
                nblocks=1,
                anomalous=anomalous,
            )
            free_Ih_table = IhTable(
                free_tables,
                space_group,
                free_indices_list,
                nblocks=1,
                anomalous=anomalous,
            )
        return global_Ih_table, free_Ih_table

    def _create_Ih_table(self):
        """Create a new Ih table from the reflection tables."""
        tables = [
            s.get_reflections_for_model_minimisation() for s in self.active_scalers
        ]
        indices_lists = [s.scaling_selection.iselection() for s in self.active_scalers]
        self._Ih_table = IhTable(
            tables,
            self.active_scalers[0].experiment.crystal.get_space_group(),
            indices_lists=indices_lists,
            nblocks=self.params.scaling_options.nproc,
            anomalous=self.params.anomalous,
        )
        for i, scaler in enumerate(self.active_scalers):
            error_model = scaler._experiment.scaling_model.error_model
            if error_model:
                self._Ih_table.update_weights(error_model, dataset_id=i)

    def make_ready_for_scaling(self, outlier=True):
        """
        Prepare the datastructures for a round of scaling.

        Update the scaling selection, create a new Ih_table and update the model
        data ready for minimisation. Also check to see if any datasets should be
        removed.
        """
        datasets_to_remove = []
        for i, scaler in enumerate(self.active_scalers):
            if outlier:
                scaler.scaling_selection = scaler.scaling_subset_sel & ~scaler.outliers
            else:
                scaler.scaling_selection = copy.deepcopy(scaler.scaling_subset_sel)
            if scaler.scaling_selection.count(True) == 0:
                datasets_to_remove.append(i)
        if datasets_to_remove:
            self.remove_datasets(self.active_scalers, datasets_to_remove)
            (
                self._global_Ih_table,
                self._free_Ih_table,
            ) = self._create_global_Ih_table(self.params.anomalous)
        self._create_Ih_table()
        self._update_model_data()

    @Subject.notify_event(event="performed_outlier_rejection")
    def round_of_outlier_rejection(self):
        self._round_of_outlier_rejection(target=None)

    def _round_of_outlier_rejection(self, target=None):
        """
        Perform a round of outlier rejection across all datasets.

        After identifying outliers, set the outliers property in individual scalers.
        """
        assert self.active_scalers is not None
        if not self.global_Ih_table:
            self._global_Ih_table, self._free_Ih_table = self._create_global_Ih_table(
                anomalous=self.params.anomalous
            )
        if self.params.scaling_options.outlier_rejection:
            if target:
                method = "target"
            else:
                method = self.params.scaling_options.outlier_rejection
            outlier_index_arrays = determine_outlier_index_arrays(
                self.global_Ih_table,
                method,
                self.params.scaling_options.outlier_zmax,
                target=target,
            )
            for outlier_indices, scaler in zip(
                outlier_index_arrays, self.active_scalers
            ):
                scaler.outliers = flex.bool(scaler.n_suitable_refl, False)
                scaler.outliers.set_selected(flumpy.from_numpy(outlier_indices), True)
            if self.params.scaling_options.emax:
                Esq_outlier_indices = determine_Esq_outlier_index_arrays(
                    self.global_Ih_table,
                    self.active_scalers[0].experiment,
                    self.params.scaling_options.emax,
                )
                for Esq_outliers, scaler in zip(
                    Esq_outlier_indices, self.active_scalers
                ):
                    scaler.outliers.set_selected(flumpy.from_numpy(Esq_outliers), True)
            if self._free_Ih_table:
                free_outlier_index_arrays = determine_outlier_index_arrays(
                    self._free_Ih_table,
                    method,
                    self.params.scaling_options.outlier_zmax,
                    target=target,
                )
                for outlier_indices, scaler in zip(
                    free_outlier_index_arrays,
                    self.active_scalers,
                ):
                    scaler.outliers.set_selected(
                        flumpy.from_numpy(outlier_indices), True
                    )
                if self.params.scaling_options.emax:
                    free_Esq_outlier_indices = determine_Esq_outlier_index_arrays(
                        self._free_Ih_table,
                        self.active_scalers[0].experiment,
                        self.params.scaling_options.emax,
                    )
                    for Esq_outlier_indices, scaler in zip(
                        free_Esq_outlier_indices,
                        self.active_scalers,
                    ):
                        scaler.outliers.set_selected(
                            flumpy.from_numpy(Esq_outlier_indices), True
                        )

        logger.debug("Finished outlier rejection.")
        log_memory_usage()

    def _select_reflections_for_scaling(self):
        # For multi dataset, auto means quasi random
        datasets_to_remove = []
        if self.params.reflection_selection.method in (
            None,
            Auto,
            "auto",
            "quasi_random",
        ):
            random_phil = self.params.reflection_selection.random
            block = self.global_Ih_table.Ih_table_blocks[0]
            m = block.group_multiplicities()
            mover1 = m[m > 1]
            total_refl_available = np.sum(mover1)
            avg_multi_overall = np.mean(mover1)
            n_datasets = len(self.active_scalers)
            total_target = max(
                avg_multi_overall * random_phil.min_groups, random_phil.min_reflections
            )

            # strategy - want ~1000 groups worth with quasi-random.
            target_for_qr = min(1000, random_phil.min_groups) * avg_multi_overall
            # ^ select this many for qr, then some from individual, then some
            # from overall.
            indices, dataset_ids = select_connected_reflections_across_datasets(
                self.global_Ih_table,
                self.active_scalers[0].experiment,
                random_phil.multi_dataset.Isigma_cutoff,
                min_total=target_for_qr,
            )

            n_connected_by_dataset = []
            total_qr_sel = indices.size()
            for i, scaler in enumerate(self.active_scalers):
                scaler.scaling_selection = flex.bool(scaler.n_suitable_refl, False)
                sel = dataset_ids == i
                indices_for_dataset = indices.select(sel)
                scaler.scaling_selection.set_selected(indices_for_dataset, True)
                n_connected_by_dataset.append(indices_for_dataset.size())

            min_cross_dataset = (total_target - total_qr_sel) / (2.0 * n_datasets)

            # first select individual dataset reflections, as we can then
            # calculate how many this procedure selects.
            total_individual_selection = 0
            total_indiv_dataset = flex.double()
            total_used_so_far = 0

            global_block = _select_groups_on_Isigma_cutoff(
                self.global_Ih_table.Ih_table_blocks[0],
                random_phil.multi_dataset.Isigma_cutoff,
            )
            for i, scaler in enumerate(self.active_scalers):
                # first set cross dataset connected reflections identified.
                # scaler.scaling_selection = flex.bool(scaler.n_suitable_refl, False)
                # now select randomly within dataset, first do preselection.
                if min_cross_dataset > 0:
                    sel = global_block.Ih_table["dataset_id"].to_numpy() == i
                    indiv_Ih_block = global_block.select(sel)
                    suitable_table = scaler.reflection_table.select(
                        scaler.suitable_refl_for_scaling_sel
                    )
                    presel = flumpy.to_numpy(
                        calculate_scaling_subset_ranges(suitable_table, self.params)
                    )
                    preselection = presel[
                        indiv_Ih_block.Ih_table["loc_indices"].to_numpy()
                    ]
                    block = indiv_Ih_block.select(preselection)
                    # have a block of individual dataset to chose from.

                    # now select multi > 1
                    n_h_over_1 = block.calc_nh() > 1
                    block = block.select(n_h_over_1)
                    if block.size:
                        n_groups = int(
                            ceil(
                                min_cross_dataset
                                / np.mean(block.group_multiplicities())
                            )
                        )
                        n_groups_in_table = block.n_groups
                        if n_groups < n_groups_in_table:
                            isel = flumpy.to_numpy(
                                flex.random_selection(n_groups_in_table, n_groups)
                            )
                            loc_indices = (
                                block.select_on_groups(isel)
                                .Ih_table["loc_indices"]
                                .to_numpy()
                            )
                            scaler.scaling_selection.set_selected(
                                flumpy.from_numpy(loc_indices), True
                            )
                        else:
                            loc_indices = block.Ih_table["loc_indices"].to_numpy()
                            scaler.scaling_selection.set_selected(
                                flumpy.from_numpy(loc_indices), True
                            )
                        n_refl = loc_indices.size
                    else:  # no refl with multiplicity > 1.
                        n_refl = 0
                else:
                    n_refl = 0
                total_individual_selection += n_refl
                total_indiv_dataset.append(n_refl)
                total_used_so_far += scaler.scaling_selection.count(True)

            # now do random selection across datasets
            target_left = total_target - total_used_so_far
            pc_used = total_used_so_far / total_refl_available
            if target_left > 0 and pc_used < 1:
                min_refl_needed = target_left * 1.05 / (1.0 - pc_used)
                # now do random sel on all
                n_groups_to_sel = int(min_refl_needed / avg_multi_overall)

                n_groups_in_table = global_block.n_groups
                if n_groups_to_sel < n_groups_in_table:
                    isel = flumpy.to_numpy(
                        flex.random_selection(n_groups_in_table, n_groups_to_sel)
                    )
                    sel_block = global_block.select_on_groups(isel)
                else:  # just use all
                    sel_block = global_block
            else:
                sel_block = global_block.select_on_groups(
                    np.array([], dtype=np.uint64).reshape((0,))
                )

            header = [
                "Dataset id",
                "reflections \nconnected to \nother datasets",
                "randomly selected \nreflections \nwithin dataset",
                "randomly selected \nreflections \nacross datasets",
                "combined number \nof reflections",
            ]
            rows = []
            total_overall = 0

            for i, scaler in enumerate(self.active_scalers):
                sel = sel_block.Ih_table["dataset_id"].to_numpy() == i
                indiv_block = sel_block.select(sel)
                loc_indices = indiv_block.Ih_table["loc_indices"].to_numpy()
                scaler.scaling_selection.set_selected(
                    flumpy.from_numpy(loc_indices), True
                )
                n = scaler.scaling_selection.count(True)
                rows.append(
                    [
                        scaler.reflection_table.experiment_identifiers().keys()[0],
                        str(n_connected_by_dataset[i]),
                        str(total_indiv_dataset[i]),
                        str(loc_indices.size),
                        str(n),
                    ]
                )
                total_overall += n
                scaler.scaling_subset_sel = copy.deepcopy(scaler.scaling_selection)
                scaler.scaling_selection &= ~scaler.outliers
                if not n:
                    datasets_to_remove.append(i)

            rows.append(
                [
                    "total",
                    str(total_qr_sel),
                    str(total_individual_selection),
                    str(sel_block.size),
                    str(total_overall),
                ]
            )
            logger.info(
                "Summary of reflections chosen for minimisation from each dataset (%s total):",
                total_overall,
            )
            logger.info(tabulate(rows, header))

        elif self.params.reflection_selection.method == "intensity_ranges":
            for i, scaler in enumerate(self.active_scalers):
                try:
                    overall_scaling_selection = calculate_scaling_subset_ranges_with_E2(
                        scaler.reflection_table, scaler.params
                    )
                except BadDatasetForScalingException:
                    datasets_to_remove.append(i)
                else:
                    scaler.scaling_selection = overall_scaling_selection.select(
                        scaler.suitable_refl_for_scaling_sel
                    )
                    if self._free_Ih_table:
                        scaler.scaling_selection.set_selected(
                            scaler.free_set_selection, False
                        )
                    scaler.scaling_subset_sel = copy.deepcopy(scaler.scaling_selection)
                    scaler.scaling_selection &= ~scaler.outliers
                    if scaler.scaling_selection.all_eq(False):
                        datasets_to_remove.append(i)

        elif self.params.reflection_selection.method == "use_all":
            block = self.global_Ih_table.Ih_table_blocks[0]
            n_mult = np.count_nonzero(block.calc_nh() > 1)
            logger.info(
                "Using all reflections (%s) for minimisation (%s with m>1)",
                block.size,
                n_mult,
            )
            for i, scaler in enumerate(self.active_scalers):
                if self._free_Ih_table:
                    scaler.scaling_selection = ~scaler.free_set_selection
                else:
                    scaler.scaling_selection = flex.bool(scaler.n_suitable_refl, True)
                scaler.scaling_subset_sel = copy.deepcopy(scaler.scaling_selection)
                scaler.scaling_selection &= ~scaler.outliers
                if scaler.scaling_selection.all_eq(False):
                    datasets_to_remove.append(i)
        elif self.params.reflection_selection.method == "random":
            # random means a random subset of groups,
            random_phil = self.params.reflection_selection.random
            block = self.global_Ih_table.Ih_table_blocks[0]
            sel_block = block.select(block.calc_nh() > 1)

            block = _select_groups_on_Isigma_cutoff(
                block, random_phil.multi_dataset.Isigma_cutoff
            )
            block = block.select(block.calc_nh() > 1)
            if not block.size:
                raise SystemExit(
                    "No groups left with multiplicity >1, scaling not possible."
                )
            avg_multi = np.mean(block.group_multiplicities())
            n_groups_to_sel = max(
                random_phil.min_groups,
                int(ceil(random_phil.min_reflections / avg_multi)),
            )

            n_groups_in_table = block.n_groups
            if n_groups_to_sel < n_groups_in_table:
                isel = flex.random_selection(n_groups_in_table, n_groups_to_sel)
                sel_block = block.select_on_groups(flumpy.to_numpy(isel))
            else:  # just use all
                sel_block = block
            logger.info(
                "Selected %s/%s groups (m>1) to use for minimisation (%s reflections)",
                min(n_groups_to_sel, n_groups_in_table),
                n_groups_in_table,
                sel_block.size,
            )
            for i, scaler in enumerate(self.active_scalers):
                sel = sel_block.Ih_table["dataset_id"].to_numpy() == i
                indiv_block = sel_block.select(sel)
                loc_indices = indiv_block.Ih_table["loc_indices"].to_numpy()
                scaler.scaling_selection = flex.bool(scaler.n_suitable_refl, False)
                scaler.scaling_selection.set_selected(
                    flumpy.from_numpy(loc_indices), True
                )
                scaler.scaling_subset_sel = copy.deepcopy(scaler.scaling_selection)
                scaler.scaling_selection &= ~scaler.outliers
                if scaler.scaling_selection.all_eq(False):
                    datasets_to_remove.append(i)
        else:
            raise ValueError("Invalid choice for 'reflection_selection.method'.")
        if datasets_to_remove:
            self.remove_datasets(self.active_scalers, datasets_to_remove)
            # Reinitialise the global datastructures for the reduced dataset.
            (
                self._global_Ih_table,
                self._free_Ih_table,
            ) = self._create_global_Ih_table(self.params.anomalous)

    def determine_shared_model_components(self):
        shared_components = set()
        for scaler in self.active_scalers:
            shared_components.add(
                scaler.experiment.scaling_model.get_shared_components()
            )
        if len(shared_components) == 1:
            return list(shared_components)[0]
        return None

    @Subject.notify_event(event="performed_error_analysis")
    def perform_error_optimisation(self, update_Ih=True):
        """Perform an optimisation of the sigma values."""
        if self.params.weighting.error_model.grouping == "combined":
            minimisation_groups = [[i for i, _ in enumerate(self.active_scalers)]]
        elif self.params.weighting.error_model.grouping == "individual":
            minimisation_groups = [[i] for i, _ in enumerate(self.active_scalers)]
        else:
            groups = self.params.weighting.error_model.error_model_group
            if not groups:
                logger.info(
                    """No error model groups defined, defaulting to combined error model optimisation"""
                )
                minimisation_groups = [[i for i, _ in enumerate(self.active_scalers)]]
            else:
                all_datasets = [i for i, _ in enumerate(self.active_scalers)]
                # groups are defined in terms of sweeps (1,2,3,...), but here
                # need to convert to dataset number (0, 1, 2,...)
                explicitly_grouped = [i - 1 for j in groups for i in j]
                if -1 in explicitly_grouped:  # sweeps provided indexed from 0
                    explicitly_grouped = [i for j in groups for i in j]
                    minimisation_groups = [list(g) for g in groups]
                else:
                    minimisation_groups = [[i - 1 for i in g] for g in groups]
                others = set(all_datasets).difference(set(explicitly_grouped))
                if others:
                    minimisation_groups += [list(others)]

        for g in minimisation_groups:
            scalers = [self.active_scalers[i] for i in g]
            error_model = scalers[0]._experiment.scaling_model.error_model
            if not error_model.params.minimisation:
                continue
            tables = [s.get_valid_reflections().select(~s.outliers) for s in scalers]
            space_group = scalers[0].experiment.crystal.get_space_group()
            Ih_table = IhTable(tables, space_group, anomalous=True)
            if len(minimisation_groups) == 1:
                logger.info("Determining a combined error model for all datasets")
            else:
                logger.info(
                    f"Error model determination for sweep(s) {','.join(str(i+1) for i in g)}"
                )
            try:
                model = run_error_model_refinement(
                    scalers[0]._experiment.scaling_model.error_model, Ih_table
                )
            except (ValueError, RuntimeError) as e:
                logger.info(e)
                logger.debug(e, exc_info=True)
            else:
                for s in scalers:
                    s._experiment.scaling_model.set_error_model(model)
        if update_Ih:
            if len(minimisation_groups) == 1:
                # one overall error model was refined
                error_model = self.active_scalers[
                    0
                ]._experiment.scaling_model.error_model
                if not error_model.params.minimisation:
                    return  # no need to update
                self.global_Ih_table.update_weights(error_model)
                if self._free_Ih_table:
                    self._free_Ih_table.update_weights(error_model)
            else:
                for i, scaler in enumerate(self.active_scalers):
                    error_model = scaler._experiment.scaling_model.error_model
                    if not error_model.params.minimisation:
                        continue  # no need to update for this subset
                    self.global_Ih_table.update_weights(error_model, dataset_id=i)
                    if self._free_Ih_table:
                        self._free_Ih_table.update_weights(error_model, dataset_id=i)


class MultiScaler(MultiScalerBase):
    """Scaler for multiple datasets where all datasets are being minimised."""

    id_ = "multi"

    def __init__(self, single_scalers):
        """
        Initialise a multiscaler from a list of single scalers.

        Create a global_Ih_table, an Ih_table to use for minimisation and update
        the data in the model components.
        """
        logger.info("Configuring a MultiScaler to handle the individual Scalers. \n")
        super().__init__(single_scalers)
        logger.info("Determining symmetry equivalent reflections across datasets.\n")
        self._active_scalers = self.single_scalers
        self._global_Ih_table, self._free_Ih_table = self._create_global_Ih_table(
            self.params.anomalous
        )
        # now select reflections from across the datasets
        self._select_reflections_for_scaling()
        self._create_Ih_table()
        # now add data to scale components from datasets
        self._update_model_data()
        logger.info("Completed configuration of MultiScaler. \n\n" + "=" * 80 + "\n")
        log_memory_usage()

    def fix_initial_parameter(self):
        for scaler in self.active_scalers:
            fixed = scaler.experiment.scaling_model.fix_initial_parameter(self.params)
            if fixed:
                return fixed

    def combine_intensities(self):
        """Combine reflection intensities, either jointly or separately."""
        if self.params.reflection_selection.combine.joint_analysis:
            logger.info(
                "Performing multi-dataset profile/summation intensity optimisation."
            )
            try:
                combiner = MultiDatasetIntensityCombiner(self)
            except DialsMergingStatisticsError as e:
                logger.info("Intensity combination failed with the error %s", e)
            else:
                for i, scaler in enumerate(self.active_scalers):
                    (
                        intensity,
                        variance,
                    ) = combiner.calculate_suitable_combined_intensities(i)
                    isel = scaler.suitable_refl_for_scaling_sel.iselection()
                    scaler.reflection_table["intensity"].set_selected(isel, intensity)
                    scaler.reflection_table["variance"].set_selected(isel, variance)
                    self.global_Ih_table.update_data_in_blocks(
                        intensity, i, column="intensity"
                    )
                    self.global_Ih_table.update_data_in_blocks(
                        variance, i, column="variance"
                    )
                    self.global_Ih_table.update_weights(
                        error_model=scaler._experiment.scaling_model.error_model,
                        dataset_id=i,
                    )
                    scaler.experiment.scaling_model.record_intensity_combination_Imid(
                        combiner.max_key
                    )
                    if self._free_Ih_table:
                        self._free_Ih_table.update_data_in_blocks(
                            intensity, i, column="intensity"
                        )
                        self._free_Ih_table.update_data_in_blocks(
                            variance, i, column="variance"
                        )
                        self._free_Ih_table.update_weights(
                            error_model=scaler._experiment.scaling_model.error_model,
                            dataset_id=i,
                        )
                self.global_Ih_table.calc_Ih()
                if self._free_Ih_table:
                    self._free_Ih_table.calc_Ih()
        else:
            for i, scaler in enumerate(self.single_scalers):
                scaler.combine_intensities()
                intensity = scaler.reflection_table["intensity"].select(
                    scaler.suitable_refl_for_scaling_sel
                )
                variance = scaler.reflection_table["variance"].select(
                    scaler.suitable_refl_for_scaling_sel
                )
                self.global_Ih_table.update_data_in_blocks(
                    intensity, i, column="intensity"
                )
                self.global_Ih_table.update_data_in_blocks(
                    variance, i, column="variance"
                )
                self.global_Ih_table.update_weights(
                    error_model=scaler._experiment.scaling_model.error_model,
                    dataset_id=i,
                )
                if self._free_Ih_table:
                    self._free_Ih_table.update_data_in_blocks(
                        intensity, i, column="intensity"
                    )
                    self._free_Ih_table.update_data_in_blocks(
                        variance, i, column="variance"
                    )
                    self._free_Ih_table.update_weights(
                        error_model=scaler._experiment.scaling_model.error_model,
                        dataset_id=i,
                    )
            self.global_Ih_table.calc_Ih()
            if self._free_Ih_table:
                self._free_Ih_table.calc_Ih()


class TargetScaler(MultiScalerBase):
    """A target scaler for scaling datasets against already scaled data."""

    id_ = "target"

    def __init__(self, scaled_scalers, unscaled_scalers):
        """
        Initialise a multiscaler from a list of single and unscaled scalers.

        First, set the active scalers (the unscaled scalers) and use these to
        create a global_Ih_table. Then, use the scaled_scalers to create a
        target_Ih_table. Create an Ih_table to use for minimisation and use
        the target_Ih_table to set the Ih_values. Finally, update the data in
        the model components.
        """
        logger.info("\nInitialising a TargetScaler instance. \n")
        super().__init__(scaled_scalers)
        logger.info("Determining symmetry equivalent reflections across datasets.\n")
        self.unscaled_scalers = unscaled_scalers
        self._active_scalers = unscaled_scalers
        self._target_active_scalers = unscaled_scalers
        self._global_Ih_table, self._free_Ih_table = self._create_global_Ih_table(
            self.params.anomalous
        )
        self._select_reflections_for_scaling()
        tables = [
            s.reflection_table.select(s.suitable_refl_for_scaling_sel).select(
                s.scaling_selection
            )
            for s in self.single_scalers
        ]
        self._target_Ih_table = IhTable(
            tables,
            self.active_scalers[0].experiment.crystal.get_space_group(),
            nblocks=1,
            anomalous=self.params.anomalous,
        )  # Keep in one table for matching below
        self._create_Ih_table()
        self._update_model_data()
        logger.info("Completed initialisation of TargetScaler. \n\n" + "=" * 80 + "\n")
        log_memory_usage()

    @property
    def target_Ih_table(self):
        """An Ih_table containing data for the target."""
        return self._target_Ih_table

    def _create_Ih_table(self):
        super()._create_Ih_table()
        for block in self._Ih_table.blocked_data_list:
            # this step reduces the number of reflections in each block
            block.match_Ih_values_to_target(self._target_Ih_table)
        self.Ih_table.generate_block_selections()

    def round_of_outlier_rejection(self):
        """Perform a round of targeted outlier rejection."""
        self._round_of_outlier_rejection(target=self._target_Ih_table)

    def update_for_minimisation(self, apm, block_id):
        """Calculate the new parameters but don't calculate a new Ih."""
        self._update_for_minimisation(apm, block_id, calc_Ih=False)

    @Subject.notify_event(event="performed_scaling")
    def perform_scaling(self, engine=None, max_iterations=None, tolerance=None):
        """Minimise the scaling model, using a fixed-Ih target."""
        for scaler in self._target_active_scalers:
            self._active_scalers = [scaler]
            self._create_Ih_table()
            self._update_model_data()
            self._perform_scaling(
                target_type=ScalingTargetFixedIH,
                engine=engine,
                max_iterations=max_iterations,
                tolerance=tolerance,
            )
        self._active_scalers = self._target_active_scalers


class NullScaler(ScalerBase):
    """A singlescaler to allow targeted scaling against calculated intensities."""

    id_ = "null"

    def __init__(self, params, experiment, reflection):
        """Set the required properties to use as a scaler for targeted scaling."""
        super().__init__(params)
        self._experiment = experiment
        self._reflection_table = reflection
        self.n_suitable_refl = self._reflection_table.size()
        self._reflection_table["inverse_scale_factor"] = flex.double(
            self.n_suitable_refl, 1.0
        )
        if "variance" not in self._reflection_table:
            self._reflection_table["variance"] = flex.double(self.n_suitable_refl, 1.0)
        self._reflection_table.set_flags(
            flex.bool(self.n_suitable_refl, False),
            self._reflection_table.flags.excluded_for_scaling,
        )
        self.suitable_refl_for_scaling_sel = flex.bool(self.n_suitable_refl, True)
        self.outliers = flex.bool(self.n_suitable_refl, False)
        self.scaling_selection = flex.bool(self.n_suitable_refl, True)
        logger.info("Target dataset contains %s reflections", self.n_suitable_refl)
        logger.info(
            "Completed preprocessing and initialisation for this dataset."
            "\n\n" + "=" * 80 + "\n"
        )

    @property
    def experiment(self):
        """Return the experiment object for the dataset"""
        return self._experiment

    @property
    def reflection_table(self):
        """Return the reflection_table object for the dataset"""
        return self._reflection_table

    @property
    def components(self):
        """Shortcut to scaling model components."""
        return self.experiment.scaling_model.components

    def expand_scales_to_all_reflections(self, caller=None, calc_cov=False):
        """Fill in abstract method, do nothing."""

    def update_for_minimisation(self, apm, block_id=0):
        """Fill in abstract method, do nothing."""


def calc_sf_variances(components, var_cov):
    """Calculate the variances of the inverse scales."""
    # note - can we do this calculation blockwise as well - takes quite a bit of memory?
    n_param = 0
    for component in components:
        n_param += components[component].n_params
        n_refl = sum(components[component].n_refl)  # should all be same
    jacobian = sparse.matrix(n_refl, n_param)
    n_cumulative_param = 0
    scales_list = []
    derivs_list = []
    for component in components:
        s, d = components[component].calculate_scales_and_derivatives(block_id=0)
        scales_list.append(s)
        derivs_list.append(d)
    for i, component in enumerate(components):
        d_block = derivs_list[i]
        n_param = components[component].n_params
        for j, component_2 in enumerate(components):
            if component_2 != component:
                d_block = row_multiply(d_block, scales_list[j])
        jacobian.assign_block(d_block, 0, n_cumulative_param)
        n_cumulative_param += n_param
    return cpp_calc_sigmasq(jacobian.transpose(), var_cov)



from __future__ import annotations

import procrunner
import pytest

from dxtbx.model.experiment_list import ExperimentListFactory
from libtbx.phil import parse

from dials.array_family import flex


@pytest.mark.xfel
def test_kapton(tmp_path, dials_data):
    """Test script for kapton correction applied to integrated data.
    Currently only testing kapton 2019 correction on rayonix-340 at LCLS
    """
    dd = dials_data("lcls_rayonix_kapton", pathlib=True)
    image_file = dd / "hit-20181213155134902.cbf"
    mask_file = dd / "mask_rayonix340mx_4x4.pickle"
    geom_file = dd / "experiments_000.json"

    # Create phil files for the two situations being tests
    #  a. without kapton
    #  b. with kapton

    stills_process_input = parse(
        """spotfinder.lookup.mask=%s\n
                          integration.lookup.mask=%s\n
                          input.reference_geometry=%s\n
                          spotfinder.filter.min_spot_size=2\n
                          spotfinder.filter.d_min=2\n
                          spotfinder.filter.d_max=18\n
                          spotfinder.threshold.dispersion.gain=0.46\n
                          spotfinder.threshold.dispersion.global_threshold=100\n
                          indexing.known_symmetry.space_group='P 21 21 21'\n
                          indexing.known_symmetry.unit_cell='41.9 75.7 102 90 90 90'\n
                          indexing.refinement_protocol.d_min_start=2\n
                          integration.debug.output=True\n
                          integration.debug.separate_files=False\n
                          integration.debug.delete_shoeboxes=True\n
                          profile.gaussian_rs.centroid_definition=com\n """
        % (mask_file, mask_file, geom_file)
    )

    kapton_input = parse(
        """ integration {
                       absorption_correction {
                         apply=True
                         algorithm=kapton_2019
                         fuller_kapton {
                           xtal_height_above_kapton_mm {
                               value=0.04
                             }
                           rotation_angle_deg {
                             value=0.55
                             }
                           kapton_half_width_mm {
                             value=0.665
                             }
                           kapton_thickness_mm {
                             value=0.025
                             }
                           smart_sigmas=True
                           }
                         }
                       }"""
    )

    with_kapton_phil = tmp_path / "params_with_kapton.phil"
    without_kapton_phil = tmp_path / "params_without_kapton.phil"

    without_kapton_phil.write_text(
        stills_process_input.as_str()
        + "output.integrated_filename=without_kapton.mpack\noutput.integrated_experiments_filename=without_kapton.expt"
    )
    with_kapton_phil.write_text(
        stills_process_input.as_str()
        + kapton_input.as_str()
        + "output.integrated_filename=with_kapton.mpack\noutput.integrated_experiments_filename=with_kapton.expt"
    )

    command_without_kapton = (
        "dials.stills_process",
        image_file,
        "params_without_kapton.phil",
    )
    command_with_kapton = (
        "dials.stills_process",
        image_file,
        "params_with_kapton.phil",
    )
    procrunner.run(command_without_kapton, working_directory=tmp_path)
    procrunner.run(command_with_kapton, working_directory=tmp_path)

    # Now compare the 2 experimental results
    # Currently just comparing the median values to get a sense of the effect if the kapton and whether it is being applied correctly
    expt_without_kapton = ExperimentListFactory.from_json_file(
        tmp_path / "without_kapton.expt", check_format=False
    )
    refl_without_kapton = flex.reflection_table.from_file(
        tmp_path / "without_kapton.mpack"
    )
    expt_with_kapton = ExperimentListFactory.from_json_file(
        tmp_path / "with_kapton.expt", check_format=False
    )
    refl_with_kapton = flex.reflection_table.from_file(tmp_path / "with_kapton.mpack")

    without_kapton_medians = []
    with_kapton_medians = []
    count = 0
    for experiments, reflections in zip(
        (expt_without_kapton, expt_with_kapton), (refl_without_kapton, refl_with_kapton)
    ):
        all_x, all_y, all_i = flex.double(), flex.double(), flex.double()
        for expt_id, experiment in enumerate(experiments):
            refls = reflections.select(reflections["id"] == expt_id)
            for panel_id, panel in enumerate(experiment.detector):
                panel_refls = refls.select(refls["panel"] == panel_id)
                x, y, z = panel_refls["xyzobs.px.value"].parts()
                for i in range(len(panel_refls)):
                    lab_x, lab_y, lab_z = panel.get_pixel_lab_coord((x[i], y[i]))
                    all_x.append(lab_x)
                    all_y.append(lab_y)
                    all_i.append(panel_refls["intensity.sum.value"][i])

        for sel in all_x <= 0, all_x > 0, all_y <= 0, all_y > 0:
            if count == 0:
                without_kapton_medians.append(flex.median(all_i.select(sel)))
            if count == 1:
                with_kapton_medians.append(flex.median(all_i.select(sel)))
        count += 1

    # Now compare results between uncorrected and corrected data

    # x < 0 where the kapton shadow is
    assert without_kapton_medians[0] < with_kapton_medians[0]
    # x > 0 where no kapton shadow present
    assert without_kapton_medians[1] == pytest.approx(with_kapton_medians[1], abs=0.1)
    # y < 0; kapton correction should average out but should be slightly higher
    assert without_kapton_medians[2] == pytest.approx(with_kapton_medians[2], abs=5.0)
    assert without_kapton_medians[2] < with_kapton_medians[2]
    # y < 0; kapton correction should average out but should be slightly higher
    assert without_kapton_medians[3] == pytest.approx(with_kapton_medians[3], abs=5.0)
    assert without_kapton_medians[3] < with_kapton_medians[3]



"""
Test refinement of multiple narrow sequences.
"""


from __future__ import annotations

from pathlib import Path

import procrunner
import pytest

from dxtbx.model.experiment_list import ExperimentListFactory
from scitbx import matrix

from dials.algorithms.refinement.engine import Journal


def test(dials_regression, tmp_path):
    data_dir = Path(dials_regression) / "refinement_test_data" / "multi_narrow_wedges"

    selection = (2, 3, 4, 5, 6, 7, 9, 11, 12, 13, 14, 17, 18, 19, 20)

    # Combine all the separate sequences

    result = procrunner.run(
        [
            "dials.combine_experiments",
            "reference_from_experiment.beam=0",
            "reference_from_experiment.goniometer=0",
            "reference_from_experiment.detector=0",
        ]
        + [
            f"experiments={data_dir}/data/sweep_%03d/experiments.json" % n
            for n in selection
        ]
        + [
            f"reflections={data_dir}/data/sweep_%03d/reflections.pickle" % n
            for n in selection
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    # Do refinement and load the results

    # turn off outlier rejection so that test takes about 4s rather than 10s
    # set close_to_spindle_cutoff to old default
    result = procrunner.run(
        [
            "dials.refine",
            "combined.expt",
            "combined.refl",
            "scan_varying=false",
            "outlier.algorithm=null",
            "close_to_spindle_cutoff=0.05",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    refined_experiments = ExperimentListFactory.from_json_file(
        tmp_path / "refined.expt", check_format=False
    )

    # Check results are as expected

    regression_experiments = ExperimentListFactory.from_json_file(
        data_dir / "regression_experiments.json", check_format=False
    )

    for e1, e2 in zip(refined_experiments, regression_experiments):
        assert e1.crystal.is_similar_to(e2.crystal)
        # FIXME need is_similar_to for detector that checks geometry
        # assert e1.detector == e2.detector
        s0_1 = matrix.col(e1.beam.get_unit_s0())
        s0_2 = matrix.col(e1.beam.get_unit_s0())
        assert s0_1.accute_angle(s0_2, deg=True) < 0.0057  # ~0.1 mrad


def test_order_invariance(dials_regression, tmp_path):
    """Check that the order that datasets are included in refinement does not
    matter"""

    data_dir = Path(dials_regression) / "refinement_test_data" / "multi_narrow_wedges"
    selection1 = (2, 3, 4, 5, 6)
    selection2 = (2, 3, 4, 6, 5)

    # First run
    result = procrunner.run(
        [
            "dials.combine_experiments",
            "reference_from_experiment.beam=0",
            "reference_from_experiment.goniometer=0",
            "reference_from_experiment.detector=0",
        ]
        + [
            f"experiments={data_dir}/data/sweep_%03d/experiments.json" % n
            for n in selection1
        ]
        + [
            f"reflections={data_dir}/data/sweep_%03d/reflections.pickle" % n
            for n in selection1
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    result = procrunner.run(
        [
            "dials.refine",
            "combined.expt",
            "combined.refl",
            "scan_varying=false",
            "outlier.algorithm=tukey",
            "history=history1.json",
            "output.experiments=refined1.expt",
            "output.reflections=refined1.refl",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    # Second run
    result = procrunner.run(
        [
            "dials.combine_experiments",
            "reference_from_experiment.beam=0",
            "reference_from_experiment.goniometer=0",
            "reference_from_experiment.detector=0",
        ]
        + [
            f"experiments={data_dir}/data/sweep_%03d/experiments.json" % n
            for n in selection2
        ]
        + [
            f"reflections={data_dir}/data/sweep_%03d/reflections.pickle" % n
            for n in selection2
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    result = procrunner.run(
        [
            "dials.refine",
            "combined.expt",
            "combined.refl",
            "scan_varying=false",
            "outlier.algorithm=tukey",
            "history=history2.json",
            "output.experiments=refined2.expt",
            "output.reflections=refined2.refl",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    # Load results
    refined_experiments1 = ExperimentListFactory.from_json_file(
        tmp_path / "refined1.expt", check_format=False
    )
    refined_experiments2 = ExperimentListFactory.from_json_file(
        tmp_path / "refined2.expt", check_format=False
    )

    history1 = Journal.from_json_file(tmp_path / "history1.json")
    history2 = Journal.from_json_file(tmp_path / "history1.json")

    # Compare RMSDs
    rmsd1 = history1["rmsd"]
    rmsd2 = history2["rmsd"]
    for a, b in zip(rmsd1, rmsd2):
        assert a == pytest.approx(b)

    # Compare crystals
    crystals1 = [exp.crystal for exp in refined_experiments1]
    crystals2 = [exp.crystal for exp in refined_experiments2[0:8]]
    crystals2.extend([exp.crystal for exp in refined_experiments2[13:16]])
    crystals2.extend([exp.crystal for exp in refined_experiments2[8:13]])
    for a, b in zip(crystals1, crystals2):
        assert a.is_similar_to(b)



from __future__ import annotations

import os

import procrunner

from dials.array_family import flex


def plausible(table):
    # Check the reflection IDs
    assert "id" in table
    assert "miller_index" in table
    assert "s1" in table
    assert "xyzcal.px" in table
    assert "xyzcal.mm" in table
    for row in table.rows():
        assert row["id"] == 0
    return True


def test_static_prediction(dials_regression, tmpdir):
    result = procrunner.run(
        [
            "dials.predict",
            os.path.join(
                dials_regression,
                "prediction_test_data",
                "experiments_scan_static_crystal.json",
            ),
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr

    table = flex.reflection_table.from_file(tmpdir / "predicted.refl")
    assert len(table) == 1996
    assert plausible(table)


def test_scan_varying_prediction(dials_regression, tmpdir):
    result = procrunner.run(
        [
            "dials.predict",
            os.path.join(
                dials_regression,
                "prediction_test_data",
                "experiments_scan_varying_crystal.json",
            ),
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr

    table = flex.reflection_table.from_file(tmpdir / "predicted.refl")
    assert len(table) == 1934
    assert plausible(table)


def test_force_static_prediction(dials_regression, tmpdir):
    result = procrunner.run(
        [
            "dials.predict",
            os.path.join(
                dials_regression,
                "prediction_test_data",
                "experiments_scan_varying_crystal.json",
            ),
            "force_static=True",
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr

    table = flex.reflection_table.from_file(tmpdir / "predicted.refl")
    assert len(table) == 1996
    assert plausible(table)



from __future__ import annotations

import os

import wx

from scitbx.matrix import col


class SBSettingsFrame(wx.MiniFrame):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        szr = wx.BoxSizer(wx.VERTICAL)
        panel = SBSettingsPanel(self)
        self.SetSizer(szr)
        szr.Add(panel, 1, wx.EXPAND)
        szr.Fit(panel)
        self.panel = panel
        self.sizer = szr
        self.Fit()
        self.Bind(wx.EVT_CLOSE, lambda evt: self.Destroy(), self)

    # XXX Could have a set_image() function instead of referring back to
    # the frame all the time?


class SBSettingsPanel(wx.Panel):
    # XXX Names: they're not really settings.  XXX Allow for setting
    # rotation, and provide a hierarchical drop-down menu to play with
    # detector, panel, sensor and ASIC.

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        # Number of decimal digits for distances.
        self.digits = 2

        # Quad translation controls
        from wx.lib.agw.floatspin import EVT_FLOATSPIN, FloatSpin

        img = self.GetParent().GetParent().pyslip.tiles.raw_image
        d = img.get_detector()
        self._quad_spinners = []
        for serial in range(4):
            fast, slow = d.hierarchy()[serial].get_origin()[0:2]
            name_quadrant = ["Q0", "Q1", "Q2", "Q3"][serial]
            box = wx.BoxSizer(wx.HORIZONTAL)

            for (name_direction, value) in [("fast", fast), ("slow", slow)]:
                name_ctrl = name_quadrant + "_" + name_direction + "_ctrl"

                spinner = FloatSpin(
                    self, digits=self.digits, name=name_ctrl, value=value
                )
                self.Bind(EVT_FLOATSPIN, self.OnUpdateQuad, spinner)

                box.Add(
                    spinner,
                    0,
                    wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
                    5,
                )
                box.Add(
                    wx.StaticText(self, label=name_quadrant + " " + name_direction),
                    0,
                    wx.ALL | wx.ALIGN_CENTER_VERTICAL,
                    5,
                )

                setattr(self, "_" + name_ctrl, spinner)
                self._quad_spinners.append(spinner)

            sizer.Add(box)

        # Spinner amount control
        box = wx.BoxSizer(wx.HORIZONTAL)
        self._spinner_amt_control = FloatSpin(
            self,
            digits=self.digits,
            name="spin_amount",
            value=1,
            min_val=0.1,
            increment=0.1,
        )
        self.Bind(EVT_FLOATSPIN, self.OnSpinAmount, self._spinner_amt_control)
        box.Add(
            self._spinner_amt_control,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        box.Add(
            wx.StaticText(self, label="Spinner increment (mm)"),
            0,
            wx.ALL | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        sizer.Add(box)

        box = wx.BoxSizer(wx.HORIZONTAL)

        btn = wx.Button(self, label="Restore metrology")
        box.Add(btn, flag=wx.ALL, border=5)
        self.Bind(wx.EVT_BUTTON, self.OnRestoreMetrology, btn)

        btn = wx.Button(self, label="Save current metrology")
        box.Add(btn, flag=wx.ALL, border=5)
        self.Bind(wx.EVT_BUTTON, self.OnSaveMetrology, btn)

        sizer.Add(box, flag=wx.ALIGN_CENTER)

        # XXX Rename to metrology tool?

    def OnRestoreMetrology(self, event):
        print("Not implemented")

    def OnSaveMetrology(self, event):
        import pycbf

        dialog = wx.FileDialog(
            self,
            defaultDir=os.curdir,
            defaultFile="quadrants.def",
            message="Save metrology file",
            style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT,
            wildcard="Phil files (*.def)|*.def",
        )
        if dialog.ShowModal() == wx.ID_OK:
            path = str(dialog.GetPath())
            if path != "":
                # The detector object of the format instance is adjusted when the quadrant calibration
                # arrows are clicked.  Sync those adjustments to a new cbf handle, drop unneeded categories
                # (categories frame specific but not metrology specific) and write the file.
                frame = self.GetParent().GetParent()
                img = frame.pyslip.tiles.raw_image
                header = img.image_set.get_format_class()(img.full_path)
                header.sync_detector_to_cbf(img.get_detector())
                cbf = header._cbf_handle
                cbf.find_category("array_data")
                cbf.remove_category()
                cbf.find_category("array_structure")
                cbf.remove_category()
                cbf.find_category("array_intensities")
                cbf.remove_category()
                cbf.find_category("diffrn_radiation")
                cbf.remove_category()
                cbf.find_category("diffrn_radiation_wavelength")
                cbf.remove_category()
                cbf.find_category("diffrn_measurement")
                cbf.remove_category()
                cbf.find_category("diffrn_scan")
                cbf.remove_category()
                cbf.find_category("diffrn_scan_frame")
                cbf.remove_category()

                cbf.write_widefile(
                    path,
                    pycbf.CBF,
                    pycbf.MIME_HEADERS | pycbf.MSG_DIGEST | pycbf.PAD_4K,
                    0,
                )

                print("Saved cbf header to", path)

    def OnUpdateQuad(self, event):
        # Get the name of the spinner and its delta, the deviation from
        # the default value.  Update the default for the next event.
        obj = event.EventObject
        name = obj.GetName()
        value = obj.GetValue()
        delta = float(value - obj.GetDefaultValue())
        obj.SetDefaultValue(value)

        # Update the frame's effective metrology parameters.
        frame = self.GetParent().GetParent()
        img = frame.pyslip.tiles.raw_image

        quads = img.get_detector().hierarchy()

        if name == "Q0_fast_ctrl":
            quad, delta = (quads[0], col((delta, 0, 0)))
        elif name == "Q0_slow_ctrl":
            quad, delta = (quads[0], col((0, delta, 0)))
        elif name == "Q1_fast_ctrl":
            quad, delta = (quads[1], col((delta, 0, 0)))
        elif name == "Q1_slow_ctrl":
            quad, delta = (quads[1], col((0, delta, 0)))
        elif name == "Q2_fast_ctrl":
            quad, delta = (quads[2], col((delta, 0, 0)))
        elif name == "Q2_slow_ctrl":
            quad, delta = (quads[2], col((0, delta, 0)))
        elif name == "Q3_fast_ctrl":
            quad, delta = (quads[3], col((delta, 0, 0)))
        elif name == "Q3_slow_ctrl":
            quad, delta = (quads[3], col((0, delta, 0)))
        else:
            raise RuntimeError("Unknown control name " + name)

        ldm = quad.get_local_d_matrix()
        fast = (ldm[0], ldm[3], ldm[6])
        slow = (ldm[1], ldm[4], ldm[7])

        orig = col((ldm[2], ldm[5], ldm[8])) + delta

        quad.set_local_frame(fast, slow, orig)

        # Update the view, trigger redraw.
        tiles = frame.pyslip.tiles
        tiles.set_image(tiles.raw_image)
        tiles.flex_image.adjust(color_scheme=tiles.current_color_scheme)

        tiles.reset_the_cache()
        tiles.tile_cache = tiles.cache[tiles.zoom_level]
        tiles.tile_list = tiles.lru[tiles.zoom_level]
        frame.pyslip.Update()

    def OnSpinAmount(self, event):
        obj = event.EventObject
        for spinner in self._quad_spinners:
            spinner.SetIncrement(obj.GetValue())



from __future__ import annotations

import random

import pytest

from dials.algorithms.indexing.basis_vector_search import fft3d, optimise


def test_optimise_basis_vectors(setup_rlp):
    random.seed(42)  # guaranteed to be random
    max_cell = 1.3 * max(setup_rlp["crystal_symmetry"].unit_cell().parameters()[:3])
    rlp = setup_rlp["rlp"]
    strategy = fft3d.FFT3D(max_cell, n_points=256)
    basis_vectors, used = strategy.find_basis_vectors(rlp)

    for v in basis_vectors:
        target = optimise.BasisVectorTarget(rlp)
        f, g = target.compute_functional_and_gradients(v)
        assert f == target.compute_functional(v)
        g_fd = _gradient_fd(target, v)
        assert list(g) == pytest.approx(g_fd, rel=1)

        minimised = optimise.BasisVectorMinimiser(rlp, v)
        # check that the minimised vectors are broadly similar to the starting vectors
        assert v.length() == pytest.approx(minimised.x.norm(), rel=2e-2)
        assert v.elems == pytest.approx(minimised.x, abs=1.2)

    optimised = optimise.optimise_basis_vectors(rlp, basis_vectors)
    assert len(optimised) == len(basis_vectors)


def _gradient_fd(target, vector, eps=1e-6):
    grads = []
    for i in range(len(vector)):
        v = list(vector)
        v[i] -= eps
        tm, _ = target.compute_functional_and_gradients(v)
        v[i] += 2 * eps
        tp, _ = target.compute_functional_and_gradients(v)
        grads.append((tp - tm) / (2 * eps))
    return grads
